# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Models to represent a workflow definition file."""

from __future__ import annotations

import itertools
from datetime import datetime
from pathlib import Path
from typing import Any, Callable, List, Optional, Sequence, Type, TypeVar, Union, cast, overload

import bashlex
from bashlex.ast import nodevisitor
from bashlex.errors import ParsingError

from renku.command.command_builder import inject
from renku.core import errors
from renku.core.interface.plan_gateway import IPlanGateway
from renku.core.interface.project_gateway import IProjectGateway
from renku.core.util import communication
from renku.core.util.datetime8601 import local_now
from renku.core.util.os import are_paths_equal, is_subpath
from renku.core.util.util import to_string
from renku.core.workflow.plan import get_latest_plan, is_plan_removed
from renku.core.workflow.run import get_valid_parameter_name, get_valid_plan_name
from renku.domain_model.project_context import project_context
from renku.domain_model.workflow.parameter import (
    CommandInput,
    CommandOutput,
    CommandParameter,
    HiddenInput,
    MappedIOStream,
    generate_parameter_name,
)
from renku.domain_model.workflow.plan import get_duplicate_arguments_names
from renku.domain_model.workflow.workflow_file import WorkflowFileCompositePlan, WorkflowFilePlan


class WorkflowFile:
    """A workflow definition file."""

    def __init__(
        self,
        path: Union[Path, str],
        steps: List[Step],
        name: str,
        description: Optional[str] = None,
        keywords: Optional[List[str]] = None,
    ):
        self.path: str = str(path)
        self.steps: List[Step] = steps
        self.name: str = name
        self.description: Optional[str] = description
        self.keywords: List[str] = [str(k) for k in keywords] if keywords else []

        self.qualified_name = name

        self.validate()
        self.set_missing_names()

    @inject.autoparams("project_gateway")
    def to_plan(self, project_gateway: IProjectGateway) -> WorkflowFileCompositePlan:
        """Convert a workflow file to a CompositePlan."""
        project_id = project_gateway.get_project().id

        workflow_file_root_plan = WorkflowFileCompositePlan(
            date_created=None,
            derived_from=None,
            description=self.description,
            id=WorkflowFileCompositePlan.generate_id(path=self.path),
            keywords=self.keywords.copy(),
            name=self.qualified_name,
            path=self.path,
            plans=[s.to_plan(project_id=project_id) for s in self.steps],
            project_id=project_id,
        )

        # NOTE: Make derivative chains and potentially re-use existing plans
        return calculate_derivatives(workflow_file_root_plan)

    def validate(self):
        """Validate a workflow file."""

        def validate_plan_name(name, kind="Workflow file"):
            valid_name = get_valid_plan_name(name)
            if valid_name != name:
                raise errors.ParseError(f"{kind} name is invalid: '{name}' (Hint: '{valid_name}' is valid)")

        def is_name_reserved(name) -> bool:
            return name in ["inputs", "outputs", "parameters"]

        def validate_step(step):
            validate_plan_name(name=step.name, kind="Step")

            for parameter in itertools.chain(step.inputs, step.outputs, step.parameters):
                if parameter.name:
                    valid_name = get_valid_parameter_name(parameter.name)
                    if valid_name != parameter.name:
                        raise errors.ParseError(
                            f"Parameter name in step '{step.name}' is invalid: '{parameter.name}' "
                            f"(Hint: '{valid_name}' is valid)"
                        )

                    if is_name_reserved(parameter.name):
                        raise errors.ParseError(
                            "Names 'inputs', 'outputs', and 'parameters' are reserved and cannot be used as parameter "
                            f"name in step '{step.name}'"
                        )

            for parameter in itertools.chain(step.inputs, step.outputs):
                if not is_subpath(parameter.path, project_context.path):
                    raise errors.UsageError(
                        "The input/output file or directory is not in the repository. Only files in the repository are "
                        "currently supported.\nHint: Specify it as a parameter instead."
                        "\n\n\t" + str(parameter.path) + "\n\n"
                    )

            duplicates = get_duplicate_arguments_names(plan=step)
            if duplicates:
                duplicates_string = ", ".join(sorted(duplicates))
                raise errors.ParseError(f"Duplicate input, output or parameter names found: {duplicates_string}")

        validate_plan_name(self.name)

        for s in self.steps:
            if s.name == self.name:
                raise errors.ParseError(f"Step '{s.name}' cannot have the same name as the workflow file")
            validate_step(s)

    def set_missing_names(self):
        """Set missing names for attributes."""

        def is_name_unique(name, step) -> bool:
            for parameter in itertools.chain(step.inputs, step.outputs, step.parameters):
                if parameter.name == name:
                    return False
            return True

        def is_name_reserved(name) -> bool:
            return name in ["inputs", "outputs", "parameters"]

        for step in self.steps:
            for parameter in itertools.chain(step.inputs, step.outputs, step.parameters):
                if not parameter.name:
                    # NOTE: Use prefix as the name if it's unique
                    if parameter.prefix:
                        valid_name = get_valid_parameter_name(parameter.prefix.strip(" -="))
                        if is_name_unique(name=valid_name, step=step) and not is_name_reserved(valid_name):
                            parameter.name = valid_name
                            continue

                    kind = parameter.__class__.__name__.lower()
                    parameter.name = generate_parameter_name(parameter=parameter, kind=kind)


class Step:
    """A single step in a workflow file."""

    def __init__(
        self,
        *,
        command: str,
        date_created: Optional[datetime] = None,
        description: Optional[str] = None,
        inputs: Optional[List[Input]] = None,
        keywords: Optional[List[str]] = None,
        name: str,
        original_command: Optional[str] = None,
        outputs: Optional[List[Output]] = None,
        parameters: Optional[List[Parameter]] = None,
        path: Union[Path, str],
        success_codes: Optional[List[int]] = None,
        workflow_file_name: str,
    ):
        self.command: str = command
        self.date_created = date_created or local_now()
        self.description: Optional[str] = description
        self.inputs: List[Input] = inputs or []
        self.keywords: List[str] = [str(k) for k in keywords] if keywords else []
        self.name: str = name
        # NOTE: Keep a copy of command since we will modify it during processing
        self.original_command = command if original_command is None else original_command
        self.outputs: List[Output] = outputs or []
        self.parameters: List[Parameter] = parameters or []
        self.path: str = str(path)
        self.success_codes: List[int] = success_codes or []

        self.qualified_name = generate_qualified_plan_name(workflow_file_name=workflow_file_name, step_name=name)

        StepCommandParser(step=self).parse_command()

    def to_plan(self, project_id: str) -> WorkflowFilePlan:
        """Convert a step to a WorkflowFilePlan."""
        id = WorkflowFilePlan.generate_id(path=self.path, name=self.name)

        # NOTE: Add the workflow file itself as a hidden dependency
        workflow_file = HiddenInput(
            default_value=self.path,
            description="Workflow file",
            id=CommandInput.generate_id(plan_id=id, name="workflow-file", postfix="workflow-file"),
            mapped_to=None,
            name="workflow-file",
            position=None,
            postfix="workflow-file",
            prefix=None,
        )

        return WorkflowFilePlan(
            command=self.command,
            date_created=self.date_created,
            derived_from=None,  # NOTE: We create the derivatives chain later
            description=self.description,
            hidden_inputs=[workflow_file],
            id=id,
            inputs=[p.to_command_input(plan_id=id, index=i) for i, p in enumerate(self.inputs, start=1)],
            keywords=self.keywords.copy(),
            name=self.qualified_name,
            outputs=[p.to_command_output(plan_id=id, index=i) for i, p in enumerate(self.outputs, start=1)],
            parameters=[p.to_command_parameter(plan_id=id, index=i) for i, p in enumerate(self.parameters, start=1)],
            path=self.path,
            project_id=project_id,
            success_codes=self.success_codes.copy(),
        )

    def __repr__(self) -> str:
        return f"<Step {self.name}: {self.command}>"


class BaseParameter:
    """Base class for Input, Output, and Parameter."""

    def __init__(
        self,
        description: Optional[str] = None,
        implicit: bool = False,
        name: Optional[str] = None,
        name_set_by_user: Optional[bool] = None,
        position: Optional[int] = None,
        prefix: Optional[str] = None,
    ):
        self.description: Optional[str] = description
        self.implicit: bool = implicit
        self.name: Optional[str] = name
        self.name_set_by_user: bool = bool(name) if name_set_by_user is None else name_set_by_user
        self.position: Optional[int] = position
        self.prefix: Optional[str] = prefix


BaseParameterType = TypeVar("BaseParameterType", bound=BaseParameter)


class Parameter(BaseParameter):
    """A parameter for a workflow file."""

    def __init__(self, value: Any, **kwargs):
        super().__init__(**kwargs)
        self.value: Any = value

        if not value:
            raise errors.ParameterError("Attribute 'value' must be set for 'Parameter' instances")

    def to_command_parameter(self, plan_id: str, index: int) -> CommandParameter:
        """Convert to a Plan parameter."""
        postfix = self.name if self.name_set_by_user else str(index)

        return CommandParameter(
            default_value=self.value,
            derived_from=None,
            description=self.description,
            id=CommandParameter.generate_id(plan_id=plan_id, name=self.name, postfix=postfix),
            name=self.name,
            name_set_by_user=self.name_set_by_user,
            position=self.position,
            postfix=postfix,
            prefix=self.prefix,
        )


class HiddenParameter(Parameter):
    """A parameter that isn't defined by user and is created by Renku."""


class BasePath(BaseParameter):
    """Base for workflow Input/Output."""

    def __init__(self, path: str, mapped_to: Optional[str] = None, persist: bool = True, **kwargs):
        super().__init__(**kwargs)
        self.mapped_to: Optional[str] = mapped_to.lower().strip() if mapped_to else None
        self.path: str = path
        self.persist: bool = persist

        if not path:
            raise errors.ParameterError(f"Attribute 'path' must be set for '{self.__class__.__name__}' instances")


class Input(BasePath):
    """An input to a workflow file."""

    def to_command_input(self, plan_id: str, index: int) -> CommandInput:
        """Convert to a Plan input."""
        postfix = self.name if self.name_set_by_user else str(index)
        return CommandInput(
            default_value=self.path,
            derived_from=None,
            description=self.description,
            encoding_format=None,
            id=CommandInput.generate_id(plan_id=plan_id, name=self.name, postfix=postfix),
            mapped_to=MappedIOStream.from_str(self.mapped_to) if self.mapped_to else None,
            name=self.name,
            name_set_by_user=self.name_set_by_user,
            position=self.position,
            postfix=postfix,
            prefix=self.prefix,
        )


class Output(BasePath):
    """An output from a workflow file."""

    def to_command_output(self, plan_id: str, index: int) -> CommandOutput:
        """Convert to a Plan output."""
        postfix = self.name if self.name_set_by_user else str(index)
        path = Path(self.path)
        create_folder = not path.exists() and path.parent.resolve() != Path.cwd()

        return CommandOutput(
            create_folder=create_folder,
            default_value=self.path,
            derived_from=None,
            description=self.description,
            encoding_format=None,
            id=CommandOutput.generate_id(plan_id=plan_id, name=self.name, postfix=postfix),
            mapped_to=MappedIOStream.from_str(self.mapped_to) if self.mapped_to else None,
            name=self.name,
            name_set_by_user=self.name_set_by_user,
            position=self.position,
            postfix=postfix,
            prefix=self.prefix,
        )


def generate_qualified_plan_name(workflow_file_name: str, step_name: str) -> str:
    """Generate name for ``WorkflowFile`` and ``Step``."""
    # NOTE: Call ``as_posix()`` to convert Windows paths to posix paths to have a platform-independent name
    return f"{workflow_file_name}.{step_name}"


@inject.autoparams("plan_gateway")
def calculate_derivatives(
    plan: Union[WorkflowFileCompositePlan, WorkflowFilePlan],
    plan_gateway: IPlanGateway,
    sequence: Optional[itertools.count] = None,
) -> Union[WorkflowFileCompositePlan, WorkflowFilePlan]:
    """Check for existing plans and set derivative chain if needed.

    Args:
        plan(Union[WorkflowFileCompositePlan, WorkflowFilePlan]): The potential derivative plan.
        plan_gateway(IPlanGateway): ``PlanGateway`` instance.
        sequence(Optional[itertools.count]): sequence is used to generate deterministic IDs in case a previous ID
             cannot be used for a plan (e.g. when a plan was deleted and re-executed) (Default value = False).

    Returns:
        Union[WorkflowFileCompositePlan, WorkflowFilePlan]: An existing plan, if found and is the same; otherwise, the
            passed-in ``plan``.
    """

    def get_next_sequence_value() -> int:
        nonlocal sequence

        if not sequence:
            sequence = itertools.count(0)

        return next(sequence)

    existing_plan = plan_gateway.get_by_id(plan.id)
    existing_plan = get_latest_plan(plan=existing_plan)

    if not existing_plan:
        # NOTE: We need to check sub-plans
        assign_new_id = False
        create_derivatives = False
    elif is_plan_removed(existing_plan):
        # NOTE: The plan was deleted. Re-call with the next sequence ID
        sequence_value = get_next_sequence_value()
        plan.assign_new_id(sequence=sequence_value)
        return calculate_derivatives(plan=plan, sequence=sequence)
    elif not isinstance(existing_plan, type(plan)):
        # NOTE: A plan with the same ID existing that is not part of a workflow file, so, we cannot form a derivation
        # chain. Re-call with the next sequence ID.
        # should be improbable. Just change the ID. This also makes it impossible to ever form a derivation chain.
        sequence_value = get_next_sequence_value()
        plan.assign_new_id(sequence=sequence_value)
        return calculate_derivatives(plan=plan, sequence=sequence)
    elif plan.is_equal_to(existing_plan):  # type: ignore
        # NOTE: It's the same plan, so, re-use the existing one (which also re-uses sub-plans if any).
        return existing_plan  # type: ignore
    else:
        assign_new_id = True
        create_derivatives = True

    if assign_new_id:
        # NOTE: No need to pass ``sequence`` here since this ID doesn't need to be stable
        plan.assign_new_id()

    if create_derivatives:
        assert existing_plan  # NOTE: To make mypy happy!
        plan.derived_from = existing_plan.id

    if isinstance(plan, WorkflowFileCompositePlan):
        # NOTE: Chain or re-use sub-plans
        steps = []
        for step in plan.plans:
            potential_existing_step = calculate_derivatives(plan=step)
            steps.append(potential_existing_step)

        plan.plans = steps

    return plan


class StepCommandParser(nodevisitor):
    """A class to parse a workflow file step's command and figure out position of its arguments."""

    def __init__(self, step: Step):
        self.step: Step = step
        self.processed_parameters: List[Union[BasePath, BaseParameter]] = []
        self.command: str = ""

    def parse_command(self) -> None:
        """Parse the command and assign some properties (e.g. position) for parameters and the command."""
        self.processed_parameters = []

        try:
            # NOTE: Parse the command to an AST
            parsed_commands = bashlex.parse(self.step.command)
        except ParsingError as e:
            raise errors.ParseError(f"Cannot parse command {self.step.command}: {e}")

        if len(parsed_commands) > 1:
            raise errors.ParseError(f"Command must be a single line: {self.step.command}")

        parsed_command = parsed_commands[0]

        # NOTE: Process each token in the AST
        self.visit(parsed_command)

        # NOTE: Set step's command
        self.step.command = self.command

        unprocessed_nodes = [self.get_node_as_string(n) for n in parsed_command.parts if not n.processed]
        if unprocessed_nodes:
            nodes_str = "\n\t".join(unprocessed_nodes)
            communication.warn(
                f"The following command tokens weren't processed in step '{self.step.name}':\n\t{nodes_str}"
            )

        parameters_with_position = (
            p
            for p in itertools.chain(self.step.inputs, self.step.outputs, self.step.parameters)
            if p.position is not None
        )
        parameters = sorted(parameters_with_position, key=lambda p: p.position)  # type: ignore
        for index, parameter in enumerate(parameters, start=1):
            parameter.position = index

        unprocessed_parameters = [
            p.name or p.prefix or str(getattr(p, "value", "")) or str(getattr(p, "path", ""))
            for p in itertools.chain(self.step.inputs, self.step.outputs, self.step.parameters)
            if not p.implicit and p not in self.processed_parameters
        ]
        if unprocessed_parameters:
            parameters_str = "\n\t".join(unprocessed_parameters)
            communication.warn(
                f"The following inputs/outputs/parameters didn't appear in the command of step '{self.step.name}': "
                f"'{self.step.original_command}':"
                f"\n(set 'implicit' attribute for them to 'true' if this is intentional)\n\t{parameters_str}"
            )

    def visitnode(self, node):
        """Start of processing a single command token."""
        if node.kind in ["list", "pipeline"]:
            raise errors.ParseError(
                f"Cannot run multiple commands/piping in step '{self.step.name}': {self.step.command}"
            )
        if node.kind not in ("command", "word", "redirect", "tilde", "parameter"):
            message = self.get_node_as_string(node)
            raise errors.ParseError(
                f"Unsupported command token: '{message}' in step '{self.step.name}': {self.step.command}"
            )

        # NOTE: Set a marker for each node if not already set
        if getattr(node, "processed", None) is None:
            node.processed = False

    def get_node_as_string(self, node) -> str:
        """Return user-friendly error message for ``bashlex`` nodes."""
        return f"Node {node.kind}: {self.step.command[node.pos[0]:node.pos[1]]}"

    def visitcommand(self, node, parts):
        """Process a full command.

        NOTE: There is exactly one command node since we don't allow piping or multiple commands. Therefore, this method
        is called only once.
        """
        if not parts:
            message = self.get_node_as_string(node)
            raise errors.ParseError(f"No command to parse in step '{self.step.name}': {message}")

        node.processed = True

        position_offset = 0

        def get_position(node, increase_offset: bool = False) -> int:
            nonlocal position_offset

            value = position_offset + node.pos[0]
            if increase_offset:
                position_offset += 1

            return value

        def process_variable(token: str, node) -> bool:
            """If token is a variable then set position for its parameter(s) and return True."""
            # NOTE: Check for $$ which is used to escape $
            if token.startswith("$$") or not token.startswith("$"):
                return False

            name = token.replace("$", "", 1)
            name = self.unescape_dollar_sign(name)
            parameter: BaseParameter

            if name == "parameters":
                for parameter in self.step.parameters:
                    parameter.position = get_position(node, increase_offset=True)
                    self.processed_parameters.append(parameter)
            elif name == "inputs":
                for parameter in self.step.inputs:
                    parameter.position = get_position(node, increase_offset=True)
                    self.processed_parameters.append(parameter)
            elif name == "outputs":
                for parameter in self.step.outputs:
                    parameter.position = get_position(node, increase_offset=True)
                    self.processed_parameters.append(parameter)
            else:
                argument = self.find_argument_by_name(name=name)
                if not argument:
                    raise errors.ParseError(f"Cannot find a parameter with name '{name}' in step '{self.step.name}'")
                argument.position = get_position(node)
                self.processed_parameters.append(argument)

            return True

        command: str = ""
        command_done: bool = False
        prefix: str = ""

        for index, node in enumerate(parts):
            # NOTE: We assume all redirections come at the end of the command. So, stop processing once reaching a
            # ``redirect`` node. Redirects are processed separately.
            if node.kind == "redirect":
                break

            if node.kind == "word":
                node.processed = True
                word = node.word
                assert isinstance(word, str), f"Token value isn't string: {node} in step '{self.step.name}'"

                if process_variable(token=word, node=node):
                    continue

                word = self.unescape_dollar_sign(word)

                # NOTE: If no command has not been found yet, don't search in the outputs
                argument = self.find_argument(
                    prefix=prefix,
                    value=word,
                    search_outputs=bool(command),
                    exclude_prefix=False,
                )

                # NOTE: Search with no ``prefix`` if we cannot find a match and command is not done yet.
                if not argument and not command_done:
                    argument = self.find_argument(
                        prefix=None,
                        value=word,
                        search_outputs=bool(command),
                        exclude_prefix=True,
                    )

                if argument:
                    command_done = True
                    prefix = ""
                    argument.position = get_position(node)
                    self.processed_parameters.append(argument)
                elif word == "--":
                    # NOTE: Command ends with we see ``--``
                    command_done = True
                    if prefix:
                        command = f"{command} {prefix}"
                        prefix = ""
                    position = get_position(node)
                    parameter = HiddenParameter(name=f"hidden-{position}", value=word, position=position)
                    self.step.parameters.append(parameter)
                    self.processed_parameters.append(parameter)
                elif not command_done and command and not prefix:
                    # NOTE: If a command is already found, word might be a prefix
                    prefix = word
                elif not command_done and command and prefix:
                    # NOTE: Previous ``word`` wasn't a prefix, so, add it to the command
                    command = f"{command} {prefix}"
                    prefix = word
                elif not command_done:
                    command = word if not command else f"{command} {word}"
                elif prefix:
                    # NOTE: The prefix was (probably) a parameter that wasn't defined
                    raise errors.ParseError(
                        f"Cannot find an argument for '{prefix}' in step '{self.step.name}': {self.step.command}"
                    )
                else:
                    prefix = word
            elif node.kind == "tilde":
                # NOTE: ``tilde`` nodes appear when there is a path with ``~`` (e.g. ~/data.csv)
                node.processed = True

        if prefix:
            if not command_done:
                command = f"{command} {prefix}"
            else:
                raise errors.ParseError(
                    f"Cannot find an argument for '{prefix}' in step '{self.step.name}': {self.step.command}"
                )

        self.command = command

    def visitredirect(self, node, input, type, output, heredoc):
        """Process a redirect."""

        def is_parameter(node):
            return (
                node.output
                and isinstance(getattr(node.output, "word", None), str)
                and node.output.word.startswith("$")
                and not node.output.word.startswith("$$")
            )

        if type in (">", ">|"):  # stdout or stderr
            if input is not None and (not isinstance(input, int) or input not in (1, 2)):
                raise errors.ParseError(
                    f"Unsupported output redirection '{input}{type}' in step '{self.step.name}': {self.step.command}"
                )

            stream = "stdout" if input is None or input == 1 else "stderr"  # input == 2
            parameter: Optional[BasePath]

            if is_parameter(node):
                name = node.output.word.replace("$", "", 1)
                name = self.unescape_dollar_sign(name)
                if name == "outputs":
                    if len(self.step.outputs) != 1:
                        raise errors.ParseError(
                            f"There must be exactly one output when using '$outputs' as {stream} "
                            f"in step '{self.step.name}'"
                        )
                    parameter = self.step.outputs[0]
                else:
                    parameter = cast(BasePath, self.find_argument_by_name(name=name, collection=self.step.outputs))
                    if not parameter:
                        raise errors.ParseError(
                            f"Cannot find variable '{name}' for {stream} in step '{self.step.name}'"
                        )
            else:
                path = output.word
                parameter = self.find_output(path=path, prefix=None, exclude_prefix=False)
                if not parameter:
                    raise errors.ParseError(
                        f"Cannot find '{stream}' == '{path}' in outputs in step '{self.step.name}': {self.step.command}"
                    )

            parameter.mapped_to = stream
        elif type == "<":  # stdin
            if is_parameter(node):
                name = node.output.word.replace("$", "", 1)
                name = self.unescape_dollar_sign(name)
                if name == "inputs":
                    if len(self.step.inputs) != 0:
                        raise errors.ParseError(
                            f"There must be exactly one input when using '$inputs' as stdin in step '{self.step.name}'"
                        )
                    parameter = self.step.inputs[0]
                else:
                    parameter = cast(BasePath, self.find_argument_by_name(name=name, collection=self.step.inputs))
                    if not parameter:
                        raise errors.ParseError(f"Cannot find variable '{name}' for stdin")
            else:
                path = output.word
                parameter = self.find_input(path=path, prefix=None, exclude_prefix=False)
                if not parameter:
                    raise errors.ParseError(
                        f"Cannot find 'stdin' == '{path}' in inputs in step '{self.step.name}': {self.step.command}"
                    )

            parameter.mapped_to = "stdin"
        elif type in (">&", "<&"):
            # NOTE: Renku metadata doesn't support a stream that is mapped to another
            stream = f"{to_string(input, strip=True)}{type}{output}"
            raise errors.ParseError(
                f"Stream-to-stream redirection '{stream}' isn't supported in step '{self.step.name}': "
                f"{self.step.command}"
            )
        elif type == ">>":
            stream = f"{to_string(input, strip=True)}{type}{output}"
            raise errors.ParseError(
                f"Appending redirection '{stream}' is not supported in step '{self.step.name}': {self.step.command}"
            )
        else:
            raise errors.ParseError(
                f"Unsupported redirection operator '{type}' in step '{self.step.name}': {self.step.command}"
            )

        node.processed = True
        node.output.processed = True

        parameter.position = node.pos[0]
        self.processed_parameters.append(parameter)

    @staticmethod
    def unescape_dollar_sign(word: str) -> str:
        """Unescape $ in the command."""
        return word.replace("$$", "$")

    def find_argument_by_name(self, name: str, collection=None) -> Optional[BaseParameter]:
        """Find an inputs/outputs/parameters with the given name."""
        collection = collection or itertools.chain(self.step.inputs, self.step.outputs, self.step.parameters)
        argument = next((p for p in collection if p.name == name), None)
        if argument:
            self.processed_parameters.append(argument)

        return argument

    def find_argument(
        self, prefix: Optional[str], value: str, search_outputs: bool, exclude_prefix: bool
    ) -> Optional[BaseParameter]:
        """Check the argument against inputs/outputs/parameters to find its type.

        Args:
            prefix(Optional[str]): Argument prefix.
            value(str): Argument value.
            search_outputs(bool): Whether to search outputs or not. When no part of a command hasn't found yet, we
                shouldn't search outputs.
            exclude_prefix(bool): Whether prefix must be excluded in search or not. When the command hasn't been fully
                found, the prefix might be part of the command and not the argument.

        Returns:
            Optional[BaseParameter]: An Input, Output, or Parameter instance if a match is found; None otherwise.
        """
        parameter = self.find_parameter(prefix=prefix, value=value, exclude_prefix=exclude_prefix)
        if parameter:
            return parameter

        input = self.find_input(prefix=prefix, path=value, exclude_prefix=exclude_prefix)
        if input:
            return input

        if search_outputs:
            output = self.find_output(prefix=prefix, path=value, exclude_prefix=exclude_prefix)
            if output:
                return output

        return None

    def find_parameter(self, prefix: Optional[str], value: str, exclude_prefix: bool) -> Optional[Parameter]:
        """Find a parameter based on its value and prefix; adjust parameter's prefix if needed."""
        return self.find_and_process_argument(
            cls=Parameter,
            collection=self.step.parameters,
            prefix=prefix,
            value=value,
            exclude_prefix=exclude_prefix,
            are_values_equal=lambda a, b: a == b,
        )

    def find_input(self, prefix: Optional[str], path: str, exclude_prefix: bool) -> Optional[BasePath]:
        """Search for a given path in inputs."""
        return self.find_and_process_argument(
            cls=Input,
            collection=self.step.inputs,
            prefix=prefix,
            value=path,
            exclude_prefix=exclude_prefix,
            are_values_equal=are_paths_equal,
        )

    def find_output(self, prefix: Optional[str], path: str, exclude_prefix: bool) -> Optional[BasePath]:
        """Search for a given path in outputs."""
        return self.find_and_process_argument(
            cls=Output,
            collection=self.step.outputs,
            prefix=prefix,
            value=path,
            exclude_prefix=exclude_prefix,
            are_values_equal=are_paths_equal,
        )

    @overload
    def find_and_process_argument(
        self,
        *,
        cls: Type[BaseParameterType],
        collection: Sequence[Parameter],
        prefix: Optional[str],
        value: str,
        exclude_prefix: bool,
        are_values_equal: Callable[[str, str], bool],
    ) -> Optional[Parameter]:
        ...

    @overload
    def find_and_process_argument(
        self,
        *,
        cls: Type[BaseParameterType],
        collection: Sequence[BasePath],
        prefix: Optional[str],
        value: str,
        exclude_prefix: bool,
        are_values_equal: Callable[[str, str], bool],
    ) -> Optional[BasePath]:
        ...

    def find_and_process_argument(
        self,
        *,
        cls: Type[BaseParameterType],
        collection: Sequence[Union[BasePath, Parameter]],
        prefix: Optional[str],
        value: str,
        exclude_prefix: bool,
        are_values_equal: Callable[[str, str], bool],
    ) -> Optional[Union[BasePath, Parameter]]:
        """Find the given input/output/parameter in the collection."""
        prefix = to_string(prefix, strip=True)
        value = to_string(value, strip=True)

        attribute = "value" if cls == Parameter else "path"

        # NOTE: Search for cases that prefix can be part of the command and not the argument
        if exclude_prefix:
            for parameter in collection:
                parameter_value = to_string(getattr(parameter, attribute), strip=True)

                if not parameter.prefix and are_values_equal(value, parameter_value):
                    return parameter
            return None

        has_equal = False
        new_prefix, new_value = "", ""
        if "=" in value and not prefix:
            has_equal = True
            new_prefix, new_value = value.split("=", maxsplit=1)
            new_prefix = new_prefix.strip()

        for parameter in collection:
            # NOTE: Process each parameter only once
            if parameter in self.processed_parameters:
                continue

            parameter_prefix = to_string(parameter.prefix, strip=True)
            parameter_value = to_string(getattr(parameter, attribute), strip=True)

            separator = "=" if prefix.endswith("=") else " "

            if prefix.strip(separator) == parameter_prefix.strip(separator) and are_values_equal(
                value, parameter_value
            ):
                if parameter.prefix:
                    # NOTE: Add a separator between prefix and value
                    parameter.prefix = parameter_prefix.strip(separator) + separator
                return parameter
            elif (
                has_equal and new_prefix == parameter_prefix.strip("=") and are_values_equal(new_value, parameter_value)
            ):
                if parameter.prefix:
                    # NOTE: Add a separator between prefix and value
                    parameter.prefix = parameter_prefix.strip("=") + "="
                return parameter

        # NOTE: Do this in a second loop since the previous comparisons have higher priority
        for parameter in collection:
            # NOTE: Process each parameter only once
            # NOTE: This only concerns parameters that have value but no prefix
            if parameter in self.processed_parameters or parameter.prefix:
                continue

            # NOTE: Use ``split`` to ignore multiple whitespaces in comparison
            parameter_value = to_string(getattr(parameter, attribute), strip=True).split()  # type: ignore
            if parameter_value == f"{prefix}{value}".split() or parameter_value == f"{prefix} {value}".split():
                return parameter

        return None



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Workflow file core logic."""

from __future__ import annotations

import itertools
from pathlib import Path
from typing import List, Optional, Union

import networkx as nx
from pydantic import validate_arguments

from renku.command.command_builder import inject
from renku.core import errors
from renku.core.interface.plan_gateway import IPlanGateway
from renku.core.plugin.workflow_file_parser import read_workflow_file
from renku.core.util import communication
from renku.core.util.os import is_subpath
from renku.core.workflow.execute import check_for_cycles, execute_workflow_graph
from renku.core.workflow.model.concrete_execution_graph import ExecutionGraph
from renku.core.workflow.model.workflow_file import WorkflowFile
from renku.domain_model.project_context import project_context
from renku.domain_model.workflow.workflow_file import WorkflowFileCompositePlan, WorkflowFilePlan


@inject.params(plan_gateway=IPlanGateway)
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def run_workflow_file(
    path: Union[Path, str],
    steps: List[str],
    dry_run: bool,
    workflow_file: Optional[WorkflowFile],
    provider: str,
    plan_gateway: IPlanGateway,
):
    """Run a workflow file."""
    from renku.command.view_model.workflow_file import WorkflowFileViewModel

    try:
        if not dry_run and not is_subpath(path=path, base=project_context.path):
            raise errors.ParameterError("Workflow file must be inside the project for execution")
    except errors.ProjectContextError:
        pass

    workflow_file = workflow_file or read_workflow_file(path=path, parser="renku")
    workflow = workflow_file.to_plan()

    # NOTE: Check workflow file name is unique
    existing_plan = plan_gateway.get_by_name(workflow.name)
    if existing_plan:
        if not isinstance(existing_plan, WorkflowFileCompositePlan) or existing_plan.path != workflow.path:
            raise errors.DuplicateWorkflowNameError(
                f"Duplicate workflow file name: Workflow '{workflow.name}' already exists."
            )

    # NOTE: Filter steps after calculating derivatives so that the root plan (that has a subset of steps) isn't
    # considered as a new version of the plan.
    selected_steps = filter_steps(workflow=workflow, steps=steps) if steps else [workflow]

    graph = ExecutionGraph(workflows=selected_steps, virtual_links=True)

    check_for_cycles(graph=graph)

    if not dry_run:
        # NOTE: We pass the non-filtered plan to be tracked; we can detect that a subset of steps were executed by
        # looking at the number of generated activities.
        execute_workflow_graph(dag=graph.workflow_graph, provider=provider, workflow_file_plan=workflow)
        dry_run_executions = []
    else:
        plans = list(nx.topological_sort(graph.workflow_graph))
        plan_step_mapping = {p: s for p in plans for s in workflow_file.steps if p.name == s.qualified_name}
        executed_steps = [plan_step_mapping[plan] for plan in plans]
        dry_run_executions = [f"Will execute step '{s.name}': {s.original_command}" for s in executed_steps]

    selected_steps_names = [s.name for s in selected_steps]
    workflow_file.steps = [s for s in workflow_file.steps if s.name in selected_steps_names]

    return WorkflowFileViewModel.from_workflow_file(workflow_file), dry_run_executions


def filter_steps(workflow: WorkflowFileCompositePlan, steps: List[str]) -> List[WorkflowFilePlan]:
    """Return a subset of workflow file steps."""
    selected_steps = set(steps)
    if len(steps) != len(selected_steps):
        communication.warn("Duplicated steps will be included only once.")

    not_found = selected_steps - {s.unqualified_name for s in workflow.plans}
    if not_found:
        not_found_str = ", ".join(not_found)
        raise errors.ParameterError(f"Cannot find steps: {not_found_str}")

    return [s for s in workflow.plans if s.unqualified_name in selected_steps]


def get_workflow_file_inputs_and_outputs(workflow_file: WorkflowFile, steps: List[str]) -> List[str]:
    """Return a list of all inputs and outputs that must be committed."""
    selected_steps = [s for s in workflow_file.steps if s.name in steps] if steps else workflow_file.steps

    return [io.path for step in selected_steps for io in itertools.chain(step.inputs, step.outputs) if io.persist]



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Plan view model."""

from __future__ import annotations

import json
from datetime import datetime, timedelta
from typing import List, Optional, Union, cast

from renku.command.view_model.agent import PersonViewModel
from renku.command.view_model.composite_plan import CompositePlanViewModel
from renku.domain_model.project_context import project_context
from renku.domain_model.workflow.composite_plan import CompositePlan
from renku.domain_model.workflow.parameter import CommandInput, CommandOutput, CommandParameter
from renku.domain_model.workflow.plan import AbstractPlan, Plan


class CommandInputViewModel:
    """View model for ``CommandInput``."""

    def __init__(
        self,
        name: str,
        default_value: str,
        description: Optional[str] = None,
        position: Optional[str] = None,
        prefix: Optional[str] = None,
        mapped_to: Optional[str] = None,
        plan_id: Optional[str] = None,
        encoding_format: Optional[str] = None,
    ):
        self.name = name
        self.description = description
        self.default_value = default_value
        self.position = position
        self.prefix = prefix
        self.mapped_to = mapped_to
        self.plan_id = plan_id
        self.type = "Input"
        self.exists = (project_context.path / self.default_value).exists()
        self.encoding_format = encoding_format

    @classmethod
    def from_input(cls, input: CommandInput, plan_id: Optional[str] = None):
        """Create view model from ``CommandInput``.

        Args:
            input(CommandInput): Command input to convert.

        Returns:
            View model for command input.
        """
        return cls(
            name=input.name,
            description=input.description,
            default_value=str(input.default_value),
            position=str(input.position) if input.position is not None else None,
            prefix=input.prefix,
            mapped_to=input.mapped_to.stream_type if input.mapped_to else None,
            plan_id=plan_id,
            encoding_format=",".join(input.encoding_format) if input.encoding_format else None,
        )


class CommandOutputViewModel:
    """View model for ``CommandOutput``."""

    def __init__(
        self,
        name: str,
        default_value: str,
        create_folder: bool,
        description: Optional[str] = None,
        position: Optional[str] = None,
        prefix: Optional[str] = None,
        mapped_to: Optional[str] = None,
        plan_id: Optional[str] = None,
        encoding_format: Optional[str] = None,
    ):
        self.name = name
        self.description = description
        self.create_folder = create_folder
        self.default_value = default_value
        self.position = position
        self.prefix = prefix
        self.mapped_to = mapped_to
        self.plan_id = plan_id
        self.type = "Output"
        self.exists = (project_context.path / self.default_value).exists()
        self.encoding_format = encoding_format

    @classmethod
    def from_output(cls, output: CommandOutput, plan_id: Optional[str] = None):
        """Create view model from ``CommandOutput``.

        Args:
            output(CommandOutput): Command output to convert.

        Returns:
            View model for command output.
        """
        return cls(
            name=output.name,
            description=output.description,
            create_folder=output.create_folder,
            default_value=str(output.default_value),
            position=str(output.position) if output.position is not None else None,
            prefix=output.prefix,
            mapped_to=output.mapped_to.stream_type if output.mapped_to else None,
            plan_id=plan_id,
            encoding_format=",".join(output.encoding_format) if output.encoding_format else None,
        )


class CommandParameterViewModel:
    """View model for ``CommandParameter``."""

    def __init__(
        self,
        name: str,
        default_value: str,
        description: Optional[str] = None,
        position: Optional[str] = None,
        prefix: Optional[str] = None,
        plan_id: Optional[str] = None,
    ):
        self.name = name
        self.description = description
        self.default_value = default_value
        self.position = position
        self.prefix = prefix
        self.plan_id = plan_id
        self.type = "Parameter"

    @classmethod
    def from_parameter(cls, parameter: CommandParameter, plan_id: Optional[str] = None):
        """Create view model from ``CommandParameter``.

        Args:
            parameter(CommandParameter): Command parameter to convert.

        Returns:
            View model for command parameter.
        """
        return cls(
            name=parameter.name,
            description=parameter.description,
            default_value=str(parameter.default_value),
            position=str(parameter.position) if parameter.position is not None else None,
            prefix=parameter.prefix,
            plan_id=plan_id,
        )


class PlanViewModel:
    """A view model for a ``Plan``."""

    def __init__(
        self,
        id: str,
        name: str,
        created: datetime,
        command: str,
        full_command: str,
        inputs: List[CommandInputViewModel],
        outputs: List[CommandOutputViewModel],
        parameters: List[CommandParameterViewModel],
        keywords: List[str],
        description: Optional[str] = None,
        success_codes: Optional[str] = None,
        annotations: Optional[str] = None,
        creators: Optional[List[PersonViewModel]] = None,
        touches_existing_files: Optional[bool] = None,
        last_executed: Optional[datetime] = None,
        number_of_executions: Optional[int] = None,
        latest: Optional[str] = None,
        duration: Optional[timedelta] = None,
    ):
        self.id = id
        self.name = name
        self.description = description
        self.created = created
        self.command = command
        self.full_command = full_command
        self.success_codes = success_codes
        self.inputs = inputs
        self.outputs = outputs
        self.parameters = parameters
        self.annotations = annotations
        self.creators = creators
        self.keywords = keywords
        self.touches_existing_files = touches_existing_files
        self.last_executed = last_executed
        self.number_of_executions = number_of_executions
        self.latest = latest
        self.type = "Plan"

        if duration is not None:
            self.duration = duration.seconds

    @classmethod
    def from_plan(cls, plan: Plan):
        """Create view model from ``Plan``.

        Args:
            plan(Plan): The plan to convert.

        Returns:
            View model for plan.
        """
        return cls(
            id=plan.id,
            name=plan.name,
            description=plan.description,
            created=plan.date_created,
            full_command=" ".join(plan.to_argv(with_streams=True)),
            command=plan.command,
            success_codes=", ".join(str(c) for c in plan.success_codes),
            inputs=[CommandInputViewModel.from_input(input, plan.id) for input in plan.inputs],
            outputs=[CommandOutputViewModel.from_output(output, plan.id) for output in plan.outputs],
            parameters=[CommandParameterViewModel.from_parameter(param, plan.id) for param in plan.parameters],
            annotations=json.dumps([{"id": a.id, "body": a.body, "source": a.source} for a in plan.annotations])
            if plan.annotations
            else None,
            creators=[PersonViewModel.from_person(p) for p in plan.creators] if plan.creators else None,
            keywords=plan.keywords,
            touches_existing_files=getattr(plan, "touches_existing_files", False),
            latest=getattr(plan, "latest", None),
            last_executed=getattr(plan, "last_executed", None),
            number_of_executions=getattr(plan, "number_of_executions", None),
            duration=getattr(plan, "duration", None),
        )


def plan_view(workflow: AbstractPlan, latest: bool = False) -> Union[CompositePlanViewModel, PlanViewModel]:
    """Convert a ``CompositePlan`` or ``Plan`` to a ``ViewModel``.

    Args:
        workflow(AbstractPlan): Plan to convert.
        latest(bool): Whether to get latest plan data.

    Returns:
        View model for converted Plan.
    """
    if isinstance(workflow, CompositePlan):
        return CompositePlanViewModel.from_composite_plan(workflow, latest=latest)
    return PlanViewModel.from_plan(cast(Plan, workflow))



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Plugin hooks for renku workflow customization."""
from typing import List, Tuple

import pluggy

from renku.domain_model.session import IHibernatingSessionProvider, ISessionProvider

hookspec = pluggy.HookspecMarker("renku")


@hookspec
def session_provider() -> Tuple[ISessionProvider, str]:  # type: ignore[empty-body]
    """Plugin Hook for ``session`` sub-command.

    Returns:
        A tuple of the plugin itself and the name of the plugin.
    """
    pass


def get_supported_session_providers() -> List[ISessionProvider]:
    """Returns the currently available interactive session providers."""
    from renku.core.plugin.pluginmanager import get_plugin_manager

    pm = get_plugin_manager()
    providers = pm.hook.session_provider()

    return sorted(providers, key=lambda p: p.priority)


def get_supported_hibernating_session_providers() -> List[IHibernatingSessionProvider]:
    """Returns the currently available interactive session providers that support hibernation."""
    providers = get_supported_session_providers()
    return [p for p in providers if isinstance(p, IHibernatingSessionProvider)]



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""External dataset provider."""

from __future__ import annotations

import urllib
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional, Tuple, cast

from renku.core import errors
from renku.core.dataset.providers.api import (
    AddProviderInterface,
    ProviderApi,
    ProviderCredentials,
    ProviderPriority,
    StorageProviderInterface,
)
from renku.core.dataset.providers.common import get_metadata
from renku.core.dataset.providers.models import DatasetAddAction
from renku.core.interface.storage import IStorage
from renku.core.util.os import get_absolute_path
from renku.core.util.urls import get_scheme
from renku.domain_model.project_context import project_context
from renku.infrastructure.storage.factory import StorageFactory

if TYPE_CHECKING:
    from renku.core.dataset.providers.models import DatasetAddMetadata
    from renku.domain_model.dataset import Dataset


class ExternalProvider(ProviderApi, StorageProviderInterface, AddProviderInterface):
    """External provider for remote filesystem."""

    priority = ProviderPriority.HIGHEST
    name = "External"
    is_remote = True

    def __init__(self, uri: str):
        super().__init__(uri=get_uri_absolute_path(uri).rstrip("/"))

    @staticmethod
    def supports(uri: str) -> bool:
        """External doesn't support any URI for addition. It's only for storage backends."""
        return False

    @staticmethod
    def supports_storage(uri: str) -> bool:
        """Whether or not this provider supports a given URI storage."""
        return get_scheme(uri) in ("file", "")

    @property
    def path(self) -> str:
        """Return External path."""
        return self.uri

    def get_metadata(
        self, uri: str, destination: Path, dataset_add_action: DatasetAddAction = DatasetAddAction.NONE, **_
    ) -> List["DatasetAddMetadata"]:
        """Get metadata of files that will be added to a dataset."""
        files = get_metadata(provider=self, uri=uri, destination=destination, dataset_add_action=dataset_add_action)
        for file in files:
            if file.url and not file.url.startswith("file:"):
                file.url = f"file://{file.url}"
                if file.based_on:
                    file.based_on.url = file.url
        return files

    def convert_to_storage_uri(self, uri: str) -> str:
        """Convert backend-specific URI to a URI that is usable by the IStorage implementation."""
        return f"file://{get_uri_absolute_path(uri=uri)}"

    def get_credentials(self) -> "ExternalCredentials":
        """Return an instance of provider's credential class."""
        return ExternalCredentials(provider=self)

    def get_storage(self, credentials: Optional["ProviderCredentials"] = None) -> "IStorage":
        """Return the storage manager for the provider."""
        external_configuration = {
            "type": "local",
        }

        if not credentials:
            credentials = self.get_credentials()

        return StorageFactory.get_storage(
            storage_scheme="file",
            provider=self,
            credentials=credentials,
            configuration=external_configuration,
        )

    def on_create(self, dataset: "Dataset") -> None:
        """Hook to perform provider-specific actions when creating a dataset."""
        storage = self.get_storage(credentials=None)

        # NOTE: The underlying rclone tool cannot tell if a directory within a External bucket exists or not
        if not storage.exists(self.uri):
            raise errors.ParameterError(f"External path '{self.path}' doesn't exists.")

        project_context.repository.add_ignored_pattern(pattern=str(dataset.get_datadir()))


class ExternalCredentials(ProviderCredentials):
    """External-specific credentials."""

    def __init__(self, provider: ExternalProvider):
        super().__init__(provider=provider)

    @staticmethod
    def get_credentials_names() -> Tuple[str, ...]:
        """Return a tuple of the required credentials for a provider."""
        return tuple()

    @property
    def provider(self) -> ExternalProvider:
        """Return the associated provider instance."""
        return cast(ExternalProvider, self._provider)

    def get_credentials_section_name(self) -> str:
        """Get section name for storing credentials.

        NOTE: This methods should be overridden by subclasses to allow multiple credentials per providers if needed.
        """
        return self.provider.uri


def get_uri_absolute_path(uri: str) -> str:
    """Return absolute path to the external directory without resolving symlinks.

    Support formats are ``file://<path>``, file:<path> or just ``<path>``.

    Args:
        uri(str): URI to get path from.

    Returns:
        str: Expanded/non-expanded URI's absolute path.
    """
    return get_absolute_path(urllib.parse.urlparse(uri).path, expand=True)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Manage project templates.

Description
~~~~~~~~~~~

Renku projects are initialized using a project template. Renku has a set of
built-in templates that you can use in your projects. These templates can be
listed by using:

Commands and options
~~~~~~~~~~~~~~~~~~~~

.. rst-class:: cli-reference-commands

.. click:: renku.ui.cli.template:template
   :prog: renku template
   :nested: full

Set a template
~~~~~~~~~~~~~~

You can change a project's template using ``renku template set`` command:

.. code-block:: console

    $ renku template set <template-id>

or use a template from a different source:

.. code-block:: console

    $ renku template set <template-id> --source <template-repo-url>

This command fails if the project already has a template. Use ``--force`` flag
to force-change the template.

.. note::

    Setting a template overwrites existing files in a project. Pass
    ``--interactive`` flag to get a prompt for selecting which files to keep or
    overwrite.


Update a project's template
~~~~~~~~~~~~~~~~~~~~~~~~~~~
A project's template can be update using:

.. code-block:: console

    $ renku template update

If an update is available, this commands updates all project's files that are
not modified locally by the project. Pass ``--interactive`` flag to select
which files to keep or overwrite.

Passing ``--dry-run`` flags shows the newest available template version and a
list of files that will be updated.

.. note::

    You can specify a template version for a project by passing a
    ``--reference`` when setting it (or when initializing a project). This
    approach only works for templates from sources other than Renku because
    Renku templates are bound to the Renku version. Note that although a
    reference can be a git tag, branch or commit SHA, it's recommended to use
    only git tags as a reference.

.. note::

    A template maintainer can disable updates for a template. In this case,
    ``renku update`` refuses to update the project. Passing ``--force`` flag
    causes Renku to update the template anyways.

.. note::

    Renku always preserve the project's Renku version that is set in the Dockerfile
    even if you overwrite the Dockerfile. The reason is that the project's metadata
    is not updated when setting/updating a template and therefore the project
    won't work with a different Renku version. To update Renku version you need
    to use ``renku migrate`` command.


Validating a template repository
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

If you are developing your own templates in a template repository, there are
some rules that templates have to follow. To assist in creating your own
templates, you can check that everything is ok with:

.. code-block:: console

    $ renku template validate

Running this inside a template repository (not in a Renku project) will
check that the manifest and individual templates are correct and follow
Renku template conventions, printing warnings or errors if something
needs to be changed.

.. cheatsheet::
   :group: Project Templates
   :command: $ renku template ls
   :description: List available Renku templates.
   :target: rp

.. cheatsheet::
   :group: Project Templates
   :command: $ renku template show <template>
   :description: Show detailed information for the given template.
   :target: rp

.. cheatsheet::
   :group: Project Templates
   :command: $ renku template update
   :description: Update the project's template if a newer version is available.
   :target: rp

.. cheatsheet::
   :group: Project Templates
   :command: $ renku template set <template>
   :description: Replace the project's template with the given template.
   :target: rp

.. cheatsheet::
   :group: Misc
   :command: $ renku template validate
   :description: Check a template repository for possible errors (useful when creating Renku templates).
   :target: rp

"""

import functools
import json
from typing import TYPE_CHECKING, Any, Dict, List

import click

import renku.ui.cli.utils.color as color
from renku.ui.cli.init import parse_parameters

if TYPE_CHECKING:
    from renku.command.view_model.template import TemplateChangeViewModel, TemplateViewModel


@click.group()
@click.pass_context
def template(ctx):
    """Template commands."""
    from renku.ui.cli import is_allowed_subcommand

    is_allowed_subcommand(ctx)


@template.command("ls")
@click.option("-s", "--source", help="Provide the templates repository url or path")
@click.option(
    "-r", "--reference", default=None, help="Specify the reference to checkout on the remote template repository"
)
@click.option("-v", "--verbose", is_flag=True, help="Show detailed description for templates and parameters")
def list_templates(source, reference, verbose):
    """Show available templates in Renku or in a template source."""
    from renku.command.template import list_templates_command
    from renku.ui.cli.utils.callback import ClickCallback

    result = (
        list_templates_command().with_communicator(ClickCallback()).build().execute(source=source, reference=reference)
    )

    _print_template_list(result.output, verbose=verbose)


@template.command("show")
@click.option("-s", "--source", help="Provide the templates repository url or path")
@click.option(
    "-r", "--reference", default=None, help="Specify the reference to checkout on the remote template repository"
)
@click.argument("template-id", required=False, default=None)
def show_template(source, reference, template_id):
    """Show detailed template information for a single template."""
    from renku.command.template import show_template_command
    from renku.ui.cli.utils.callback import ClickCallback

    result = (
        show_template_command()
        .with_communicator(ClickCallback())
        .build()
        .execute(source=source, reference=reference, id=template_id)
    )
    _print_template(result.output)


@template.command("set")
@click.option("-s", "--source", help="Provide the templates repository url or path")
@click.option(
    "-r", "--reference", default=None, help="Specify the reference to checkout on the remote template repository"
)
@click.option(
    "-p",
    "--parameter",
    "parameters",
    multiple=True,
    type=click.STRING,
    callback=parse_parameters,
    help=(
        "Provide parameters value. Should be invoked once per parameter. "
        'Please specify the values as follow: --parameter "param1"="value"'
    ),
)
@click.option("-f", "--force", is_flag=True, help="Override existing template")
@click.option("-i", "--interactive", is_flag=True, help="Ask for overwriting files or parameter values")
@click.option("-n", "--dry-run", is_flag=True, help="Show what would have been changed")
@click.argument("template-id", required=False, default=None)
def set_template(source, reference, template_id, parameters, force, interactive, dry_run):
    """Set a template for a project."""
    from renku.command.template import set_template_command
    from renku.ui.cli.utils.callback import ClickCallback

    result = (
        set_template_command()
        .with_communicator(ClickCallback())
        .build()
        .execute(
            source=source,
            reference=reference,
            id=template_id,
            input_parameters=parameters,
            force=force,
            interactive=interactive,
            dry_run=dry_run,
        )
    )

    if dry_run:
        _print_template_change(result.output)


@template.command("update")
@click.option("-f", "--force", is_flag=True, help="Force an update for fixed template versions")
@click.option("-i", "--interactive", is_flag=True, help="Ask for overwriting files or parameter values")
@click.option("-n", "--dry-run", is_flag=True, help="Show what would have been updated")
def update_template(force, interactive, dry_run):
    """Update a project's template."""
    from renku.command.template import update_template_command
    from renku.ui.cli.utils.callback import ClickCallback

    result = (
        update_template_command()
        .with_communicator(ClickCallback())
        .build()
        .execute(force=force, interactive=interactive, dry_run=dry_run)
    )

    if not result.output:
        click.secho("Template is up-to-date", fg="green")
    elif dry_run:
        _print_template_change(result.output)


@template.command("validate")
@click.pass_context
@click.option("json_format", "--json", is_flag=True, help="Return result as JSON.")
@click.option(
    "-r",
    "--reference",
    type=click.STRING,
    help="Git revision/branch/tag to validate the template at.",
)
@click.option(
    "-s",
    "--source",
    type=click.STRING,
    help="Remote template repository to clone and check.",
)
def validate_template(ctx, json_format, reference, source):
    """Validate a template repository and check for common issues."""
    from renku.command.template import validate_templates_command

    result = validate_templates_command().build().execute(source=source, reference=reference).output

    if json_format:
        click.echo(json.dumps(result))
    else:
        if result["warnings"]:
            click.secho("Manifest Warnings:", fg="yellow")
            for warning in result["warnings"]:
                click.secho(f"\t{warning}", fg="yellow")
        if result["manifest"]:
            click.secho(f"Manifest Errors:\n\t{result['manifest']}", fg="red")
        if result["templates"]:
            click.secho("Template Errors:", fg="red")

            for template_id, messages in result["templates"].items():
                click.secho(f"\t{template_id}:", fg="red")

                for message in messages:
                    click.secho(f"\t\t{message}", fg="red")
        if result["valid"]:
            click.secho("OK", fg=color.GREEN)
    if not result["valid"]:
        ctx.exit(1)


def _print_template(template: "TemplateViewModel"):
    """Print detailed template info."""
    from renku.core.util.util import to_string

    print_name = functools.partial(click.style, bold=True, fg="magenta")
    print_value = functools.partial(click.style, bold=True)

    def print_template_parameters():
        for p in template.parameters:
            click.echo(f"  {print_value(p.name)}:")
            click.echo(print_name("    Description: ") + print_value(p.description))
            if p.type:
                click.echo(print_name("    Type: ") + print_value(p.type))
            if p.possible_values:
                click.echo(print_name("    Possible values: ") + print_value(p.possible_values))
            if p.default is not None:
                click.echo(print_name("    Default value: ") + print_value(p.default))

    click.echo(print_name("Id: ") + print_value(template.id))
    click.echo(print_name("Id aliases: ") + print_value(template.aliases))
    click.echo(print_name("Name: ") + print_value(template.name))
    click.echo(print_name("Source: ") + print_value(template.source))
    click.echo(print_name("Reference: ") + print_value(to_string(template.reference)))
    click.echo(print_name("Version: ") + print_value(template.version))
    click.echo(print_name("Description: ") + print_value(template.description))
    click.echo(print_name("Parameters:"))
    print_template_parameters()
    click.echo(print_name("Immutable files: ") + print_value(template.immutable_files))
    click.echo(print_name("Available versions: ") + print_value(template.versions))


def _print_template_list(templates: List["TemplateViewModel"], verbose: bool):
    """Print a list of templates."""
    from renku.core.util.tabulate import tabulate

    for index, template in enumerate(templates, start=1):
        setattr(template, "index", index)

    headers: Dict[str, Any] = {"index": None, "id": None}
    if verbose:
        headers["description"] = "Description"
        # "variables": "Parameters"

    table = tabulate(templates, headers=headers)
    click.echo(table)


def _print_template_change(changes: "TemplateChangeViewModel"):
    """Print detailed template info."""
    from renku.core.util.util import to_string

    print_name = functools.partial(click.style, bold=True, fg="magenta")
    print_value = functools.partial(click.style, bold=True)

    click.echo(print_name("Id: ") + print_value(changes.id))
    click.echo(print_name("Old Id: ") + print_value(changes.old_id))
    click.echo(print_name("Source: ") + print_value(changes.source))
    click.echo(print_name("New reference: ") + print_value(to_string(changes.reference)))
    click.echo(print_name("New version: ") + print_value(to_string(changes.version)))
    click.echo(print_name("File changes:"))
    click.echo("  " + "\n  ".join(changes.file_changes))



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configuration management."""

import configparser
import os
from io import StringIO

from renku.core.constant import DATA_DIR_CONFIG_KEY
from renku.domain_model.enums import ConfigFilter
from renku.domain_model.project_context import project_context


def global_config_read_lock():
    """Create a user-level config read lock."""
    from renku.core.util.contexts import Lock

    return Lock(project_context.global_config_path)


def global_config_write_lock():
    """Create a user-level config write lock."""
    from renku.core.util.contexts import Lock

    return Lock(project_context.global_config_path, mode="exclusive")


def get_value(section, key, config_filter=ConfigFilter.ALL):
    """Get value from specified section and key."""
    config = load_config(config_filter=config_filter)
    return config.get(section, key, fallback=None)


def set_value(section, key, value, global_only=False):
    """Set value to specified section and key."""
    config_filter = ConfigFilter.GLOBAL_ONLY

    if not global_only:
        config_filter = ConfigFilter.LOCAL_ONLY
        _check_config_is_not_readonly(section, key)

    config = load_config(config_filter=config_filter)
    if section in config:
        config[section][key] = value
    else:
        config[section] = {key: value}

    store_config(config, global_only=global_only)


def remove_value(section, key, global_only=False):
    """Remove key from specified section or remove sections."""
    config_filter = ConfigFilter.GLOBAL_ONLY

    if not global_only:
        config_filter = ConfigFilter.LOCAL_ONLY
        _check_config_is_not_readonly(section, key)

    value = None
    config = load_config(config_filter=config_filter)
    removed_sections = []
    if section in config:
        if key == "*":
            value = config.pop(section)
        else:
            value = config[section].pop(key, None)

            if not config[section].keys():
                removed_sections.append(section)
    elif section.endswith("*"):
        section_prefix = section[:-1]
        for section in config:
            if section.startswith(section_prefix):
                value = config[section]
                removed_sections.append(section)

    for section in removed_sections:
        config.pop(section)
    if value is not None:
        store_config(config, global_only=global_only)
    return value


def load_config(config_filter=ConfigFilter.ALL):
    """Loads local, global or both configuration object."""
    try:
        import importlib_resources  # type:ignore
    except ImportError:
        import importlib.resources as importlib_resources  # type:ignore

    # NOTE: Use RawConfigParser because ConfigParser does non-standard INI interpolation of some values
    config = configparser.RawConfigParser()
    ref = importlib_resources.files("renku.data") / "defaults.ini"
    with importlib_resources.as_file(ref) as default_ini:
        config_files = [default_ini]

    if config_filter == ConfigFilter.LOCAL_ONLY:
        config_files += [project_context.local_config_path]
    elif config_filter == ConfigFilter.GLOBAL_ONLY:
        config_files += [project_context.global_config_path]
    elif config_filter == ConfigFilter.ALL:
        config_files += [project_context.global_config_path, project_context.local_config_path]

    if config_filter != ConfigFilter.LOCAL_ONLY:
        with global_config_read_lock():
            config.read(config_files)
    else:
        config.read(config_files)

    # NOTE: transform config section for backwards compatibility. Changes section names like
    # 'renku "interactive"' to just 'interactive' to be in line with python config conventions.
    for section in config.sections():
        if not section.startswith('renku "'):
            continue

        config[section[7:-1]] = dict(config.items(section))  # NOTE: Drop first 7 and last char
        config.pop(section)

    return config


def store_config(config, global_only):
    """Persists locally or globally configuration object.

    Global configuration is updated only when :global_only: is True,
    otherwise, updates are written to local project configuration
    """
    filepath = project_context.global_config_path if global_only else project_context.local_config_path

    if global_only:
        with global_config_write_lock():
            os.umask(0)
            fd = os.open(filepath, os.O_CREAT | os.O_RDWR | os.O_TRUNC, 0o600)
            write_config(fd, config)
    else:
        write_config(filepath, config)


def get_config(config_filter=ConfigFilter.ALL, as_string=True):
    """Read all configurations."""
    config = load_config(config_filter=config_filter)
    if as_string:
        with StringIO() as output:
            config.write(output)
            return output.getvalue()
    else:
        return {f"{s}.{k}": v for s in config.sections() for k, v in config.items(s)}


def _check_config_is_not_readonly(section, key):
    from renku.core import errors

    readonly_configs = {"renku": [DATA_DIR_CONFIG_KEY]}

    value = get_value(section, key, config_filter=ConfigFilter.LOCAL_ONLY)
    if not value:
        return

    if key in readonly_configs.get(section, []):
        raise errors.ParameterError(f"Configuration {key} cannot be modified.")


def write_config(filepath, config):
    """Write config value to a specified path."""
    with open(filepath, "w+") as file:
        config.write(file)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku core constants."""

import os
from enum import IntEnum
from pathlib import Path

FILESYSTEM_ROOT = os.path.abspath(os.sep)
"""Path to the root of the filesystem."""

APP_NAME = "Renku"
"""Application name for storing configuration."""

RENKU_HOME = ".renku"
"""Project directory name."""

RENKU_TMP = "tmp"

CACHE = "cache"
"""Directory to cache transient data."""

CONFIG_NAME = "renku.ini"
CONFIG_LOCAL_PATH = os.path.join(RENKU_HOME, CONFIG_NAME)

DATA_DIR_CONFIG_KEY = "data_directory"

DATABASE_PATH: str = "metadata"
"""Directory for metadata storage."""

DATASET_IMAGES = "dataset_images"
"""Directory for dataset images."""

IMAGES = "images"
"""Path for images/icons."""

DEFAULT_DATA_DIR = "data"

DOCKERFILE = "Dockerfile"
"""Name of the Dockerfile in the repository."""

LOCK_SUFFIX = ".lock"
"""Default suffix for Renku lock file."""

POINTERS = "pointers"
"""Directory for storing external pointer files."""

REFS = "refs"
"""Define a name of the folder with references in the Renku folder."""

RENKU_LFS_IGNORE_PATH = ".renkulfsignore"
""".gitignore like file specifying paths that are not tracked in LFS."""

TEMPLATE_CHECKSUMS = "template_checksums.json"

RENKU_PROTECTED_PATHS = [
    ".dockerignore",
    ".git",
    ".git/*",
    ".gitattributes",
    ".gitignore",
    ".gitlab-ci.yml",
    ".renku",
    ".renku/**",
    ".renkulfsignore",
    "Dockerfile*",
    "environment.yml",
    "requirements.txt",
]

DATABASE_METADATA_PATH = [
    Path(RENKU_HOME) / DATABASE_PATH,
]

PROJECT_METADATA_PATH = [
    Path(RENKU_HOME) / DATABASE_PATH,
    Path(RENKU_HOME) / IMAGES,
]

DATASET_METADATA_PATHS = [
    Path(RENKU_HOME) / DATABASE_PATH,
    Path(RENKU_HOME) / DATASET_IMAGES,
    Path(RENKU_HOME) / POINTERS,
    Path(RENKU_HOME) / REFS,
    ".gitattributes",
    ".gitignore",
]


class ProviderPriority(IntEnum):
    """Defines the order in which providers are checked (highest order is checked first).

    For example, to check if a dataset provider supports a URI, providers that support more specific URIs should have a
    higher priority so that they are checked first.
    """

    HIGHEST = 1
    HIGHER = 2
    HIGH = 3
    NORMAL = 4
    LOW = 5
    LOWER = 6
    LOWEST = 7



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Converter for workflows to CWL."""

import os
import re
import tempfile
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union, cast
from uuid import uuid4

import cwl_utils.parser.cwl_v1_2 as cwl

from renku.core import errors
from renku.core.plugin import hookimpl
from renku.core.plugin.provider import RENKU_ENV_PREFIX
from renku.core.util.yaml import dumps_yaml, write_yaml
from renku.core.workflow.model.concrete_execution_graph import ExecutionGraph
from renku.domain_model.workflow.composite_plan import CompositePlan
from renku.domain_model.workflow.converters import IWorkflowConverter
from renku.domain_model.workflow.parameter import DIRECTORY_MIME_TYPE, CommandInput, CommandOutput, CommandParameter
from renku.domain_model.workflow.plan import Plan


class CommandLineTool(cwl.CommandLineTool):
    """CommandLineTool that supports empty outputs."""

    def get_dict(self):
        """Set outputs to empty if not set."""
        d = super().get_dict()  # type: ignore[misc]
        if "outputs" not in d:
            d["outputs"] = []
        return d


class WorkflowStep(cwl.WorkflowStep):
    """WorkflowStep that supports empty outputs."""

    def get_dict(self):
        """Set out to empty if not set."""
        d = super().get_dict()  # type: ignore[misc]
        if "out" not in d:
            d["out"] = []
        return d


def _get_argument_type(value):
    """Get the type of a command line argument."""
    type_ = "string"

    if isinstance(value, float):
        type_ = "float"
    elif isinstance(value, int):
        type_ = "int"
    else:
        try:
            casted_value = float(value)

            if "." not in value:
                type_ = "int"
                casted_value = int(value)
            else:
                type_ = "float"

            value = casted_value
        except ValueError:
            pass

    return value, type_


class CWLExporter(IWorkflowConverter):
    """Converts a ``CompositePlan`` or a ``Plan`` to CWL format."""

    @hookimpl
    def workflow_format(self):
        """Workflow format name."""
        return (self, ["cwl"])

    @hookimpl
    def workflow_convert(
        self,
        workflow: Union[CompositePlan, Plan],
        basedir: Path,
        output: Optional[Path],
        output_format: Optional[str],
        resolve_paths: Optional[bool],
        nest_workflows: Optional[bool],
    ) -> str:
        """Converts the specified workflow to CWL format.

        Args:
            worflow(Union[CompositePlan, Plan]): The plan or composite plan to be converted to cwl.
            basedir(Path): The path of the base location used as a prefix for all workflow input and outputs.
            output(Optional[Path]): The file where the CWL specification should be saved,
                if None then no file is created.
            output_format(Optional[str]): Not used. Only YAML is generated, regardless of what is provided.
            resolve_paths(Optional[bool]): Whether to make all paths absolute and resolve all symlinks,
                True by default.
            nest_workflows(Optional[bool]): Whether nested CWL workflows should be used or each sub-workflow should be
                a separate file, False by default.

        Returns:
            The contents of the CWL workflow as string. If nested workflows are used then only the parent
            specification is returned.
        """
        filename = None

        if resolve_paths is None:
            resolve_paths = True

        if output:
            if output.is_dir():
                tmpdir = output
                filename = None
            else:
                tmpdir = output.parent
                filename = output
        else:
            tmpdir = Path(tempfile.mkdtemp())

        cwl_workflow: Union[cwl.Workflow, CommandLineTool]
        if isinstance(workflow, CompositePlan):
            cwl_workflow = CWLExporter._convert_composite(workflow, basedir, resolve_paths=resolve_paths)
            if nest_workflows:
                # INFO: There is only one parent workflow with all children embedded in it
                if cwl_workflow.requirements is None:
                    cwl_workflow.requirements = []
                cwl_workflow.requirements.append(cwl.SubworkflowFeatureRequirement())
            else:
                # INFO: The parent composite worfklow references other workflow files,
                # write the child workflows in separate files and reference them in parent
                for step in cast(List[WorkflowStep], cwl_workflow.steps):
                    step_filename = Path(f"{uuid4()}.cwl")
                    step_path = (tmpdir / step_filename).resolve()
                    write_yaml(step_path, step.run.save())
                    step.run = f"file://{step_path}"
            if filename is None:
                filename = Path(f"parent_{uuid4()}.cwl")
        else:
            cwl_workflow = CWLExporter._convert_step(workflow, basedir, resolve_paths=resolve_paths)
            if filename is None:
                filename = Path(f"{uuid4()}.cwl")

        cwl_workflow_dict: Dict[str, Any] = cwl_workflow.save()
        path = (tmpdir / filename).resolve()
        write_yaml(path, cwl_workflow_dict)
        return dumps_yaml(cwl_workflow_dict)

    @staticmethod
    def _sanitize_id(id):
        return re.sub(r"/|-", "_", id)

    @staticmethod
    def _convert_composite(workflow: CompositePlan, basedir: Path, resolve_paths: bool) -> cwl.Workflow:
        """Converts a composite plan to a CWL file."""
        inputs: Dict[str, str] = {}
        arguments = {}
        outputs: Dict[str, Tuple[str, Any]] = {}
        consumed_outputs = set()
        steps = []

        input_index = 1
        argument_index = 1

        graph = ExecutionGraph([workflow], virtual_links=True)
        cycles = graph.cycles
        if cycles:
            cycles = [list(map(lambda x: x.name, cycle)) for cycle in cycles]
            raise errors.GraphCycleError(cycles)

        import networkx as nx

        for i, wf in enumerate(nx.topological_sort(graph.workflow_graph)):
            step_clitool = CWLExporter._convert_step(workflow=wf, basedir=basedir, resolve_paths=resolve_paths)
            step = WorkflowStep(in_=[], out=[], run=step_clitool, id=f"step_{i}")

            for input in wf.inputs:
                input_path = input.actual_value

                sanitized_id = CWLExporter._sanitize_id(input.id)
                if input_path in inputs:
                    # already used as top-level input elsewhere, reuse
                    step.in_.append(cwl.WorkflowStepInput(sanitized_id, source=inputs[input_path]))
                elif input_path in outputs:
                    # output of a previous step, refer to it
                    consumed_outputs.add(outputs[input_path][0])
                    step.in_.append(
                        cwl.WorkflowStepInput(sanitized_id, source=f"{outputs[input_path][1]}/{outputs[input_path][0]}")
                    )
                else:
                    # input isn't output and doesn't exist yet, add new
                    inputs[input_path] = f"input_{input_index}"
                    step.in_.append(cwl.WorkflowStepInput(sanitized_id, source=inputs[input_path]))
                    input_index += 1

            for parameter in wf.parameters:
                argument_id = f"argument_{argument_index}"
                arguments[argument_id] = parameter.actual_value
                step.in_.append(cwl.WorkflowStepInput(CWLExporter._sanitize_id(parameter.id), source=argument_id))
                argument_index += 1

            for output in wf.outputs:
                sanitized_id = CWLExporter._sanitize_id(output.id)

                if output.mapped_to:
                    sanitized_id = f"output_{output.mapped_to.stream_type}"
                outputs[output.actual_value] = (sanitized_id, step.id)
                step.out.append(cwl.WorkflowStepOutput(sanitized_id))

            steps.append(step)

        workflow_object = cwl.Workflow([], [], steps, id=str(uuid4()), requirements=[], hints=[], cwlVersion="v1.0")

        # check types of paths and add as top level inputs/outputs
        for path, id_ in inputs.items():
            type_ = "Directory" if os.path.isdir(path) else "File"
            location = Path(path)
            if resolve_paths:
                location = location.resolve()
                location_str = str(location.as_uri())
            else:
                location_str = str(location)
            workflow_object.inputs.append(
                cwl.WorkflowInputParameter(
                    id=id_,
                    type=type_,
                    default={"location": location_str, "class": type_},
                )
            )

        for id_, value in arguments.items():
            value, type_ = _get_argument_type(value)
            workflow_object.inputs.append(cwl.WorkflowInputParameter(id=id_, type=type_, default=value))

        for index, (path, (id_, step_id)) in enumerate(outputs.items(), 1):
            type_ = "Directory" if os.path.isdir(path) else "File"
            workflow_object.outputs.append(
                cwl.WorkflowOutputParameter(id=f"output_{index}", outputSource=f"{step_id}/{id_}", type=type_)
            )

        return workflow_object

    @staticmethod
    def _convert_step(workflow: Plan, basedir: Path, resolve_paths: bool) -> CommandLineTool:
        """Converts a single workflow step to a CWL CommandLineTool."""
        stdin, stdout, stderr = None, None, None

        inputs = list(workflow.inputs)

        for output_ in workflow.outputs:
            if not output_.mapped_to:
                continue
            if output_.mapped_to.stream_type == "stderr":
                stderr = output_.actual_value
            if output_.mapped_to.stream_type == "stdout":
                stdout = output_.actual_value

        tool_object = CommandLineTool(
            inputs=[],
            outputs=[],
            id=str(uuid4()),
            baseCommand=workflow.command.split(" "),
            requirements=[],
            stdin=stdin,
            stderr=stderr,
            stdout=stdout,
            cwlVersion="v1.0",
        )

        workdir_req = cwl.InitialWorkDirRequirement([])
        jsrequirement = False

        environment_variables = []
        dirents = []

        for output_ in workflow.outputs:
            path = output_.actual_value
            if not os.path.isdir(path):
                path = str(Path(path).parent)
            if path != "." and path not in dirents and output_.create_folder:
                # workflow needs to create subdirectory for output file,
                # if the directory was not already present
                workdir_req.listing.append(
                    cwl.Dirent(entry='$({"listing": [], "class": "Directory"})', entryname=path, writable=True)
                )
                dirents.append(path)
                jsrequirement = True

            environment_variables.append(
                cwl.EnvironmentDef(f"{RENKU_ENV_PREFIX}{output_.name}", str(output_.actual_value))
            )
            outp, arg = CWLExporter._convert_output(output_)
            tool_object.outputs.append(outp)
            if arg:
                tool_object.inputs.append(arg)

        for input_ in inputs:
            tool_input = CWLExporter._convert_input(input_, basedir, resolve_paths=resolve_paths)

            workdir_req.listing.append(
                cwl.Dirent(entry=f"$(inputs.{tool_input.id})", entryname=input_.actual_value, writable=False)
            )

            environment_variables.append(
                cwl.EnvironmentDef(f"{RENKU_ENV_PREFIX}{input_.name}", str(input_.actual_value))
            )
            tool_object.inputs.append(tool_input)
            if input_.mapped_to:
                tool_object.stdin = f"$(inputs.{tool_input.id}.path)"
                jsrequirement = True

        for parameter in workflow.parameters:
            environment_variables.append(
                cwl.EnvironmentDef(f"{RENKU_ENV_PREFIX}{parameter.name}", str(parameter.actual_value))
            )
            tool_object.inputs.append(CWLExporter._convert_parameter(parameter))

        workdir_req.listing.append(
            cwl.Dirent(entry="$(inputs.input_renku_metadata)", entryname=".renku", writable=False)
        )
        location = basedir / ".renku"
        if resolve_paths:
            location = location.resolve()
            location_str = location.as_uri()
        else:
            location_str = str(location)
        tool_object.inputs.append(
            cwl.CommandInputParameter(
                id="input_renku_metadata",
                type="Directory",
                inputBinding=None,
                default={"location": location_str, "class": "Directory"},
            )
        )

        if workdir_req.listing:
            tool_object.requirements.append(workdir_req)  # type: ignore
        if jsrequirement:
            tool_object.requirements.append(cwl.InlineJavascriptRequirement())  # type: ignore
        if environment_variables:
            tool_object.requirements.append(cwl.EnvVarRequirement(environment_variables))  # type: ignore

        return tool_object

    @staticmethod
    def _convert_parameter(parameter: CommandParameter):
        """Converts an parameter to a CWL input."""
        value, type_ = _get_argument_type(parameter.actual_value)

        separate = None
        prefix = None
        if parameter.prefix:
            prefix = parameter.prefix
            separate = False

            if prefix.endswith(" "):
                prefix = prefix[:-1]
                separate = True

        return cwl.CommandInputParameter(
            id=CWLExporter._sanitize_id(parameter.id),
            type=type_,
            inputBinding=cwl.CommandLineBinding(position=parameter.position, prefix=prefix, separate=separate)
            if parameter.position or prefix
            else None,
            default=value,
        )

    @staticmethod
    def _convert_input(input: CommandInput, basedir: Path, resolve_paths: bool):
        """Converts an input to a CWL input."""
        type_ = (
            "Directory"
            if input.encoding_format is not None and DIRECTORY_MIME_TYPE in input.encoding_format
            else "File"
        )
        position = input.position

        sanitized_id = CWLExporter._sanitize_id(input.id)
        if input.mapped_to:
            sanitized_id = "input_stdin"
            position = None

        separate = None
        prefix = None
        if input.prefix:
            prefix = input.prefix
            separate = False

            if prefix.endswith(" "):
                prefix = prefix[:-1]
                separate = True

        location = basedir / input.actual_value
        if resolve_paths:
            location = location.resolve()
            location_str = location.as_uri()
        else:
            location_str = str(location)
        return cwl.CommandInputParameter(
            id=sanitized_id,
            type=type_,
            inputBinding=cwl.CommandLineBinding(position=position, prefix=prefix, separate=separate)
            if position or prefix
            else None,
            default={"location": location_str, "class": type_},
        )

    @staticmethod
    def _convert_output(output: CommandOutput):
        """Converts an output to a CWL output."""
        if output.mapped_to:
            return (
                cwl.CommandOutputParameter(
                    id=f"output_{output.mapped_to.stream_type}",
                    type=output.mapped_to.stream_type,
                    streamable=False,
                ),
                None,
            )

        type_ = (
            "Directory"
            if output.encoding_format is not None and DIRECTORY_MIME_TYPE in output.encoding_format
            else "File"
        )

        sanitized_id = CWLExporter._sanitize_id(output.id)

        if output.position:
            # output is specified as a parameter, create an input as well
            separate = None
            prefix = None
            if output.prefix:
                prefix = output.prefix
                separate = False

                if prefix.endswith(" "):
                    prefix = prefix[:-1]
                    separate = True

            arg = cwl.CommandInputParameter(
                id=f"{sanitized_id}_arg",
                type="string",
                inputBinding=cwl.CommandLineBinding(position=output.position, prefix=prefix, separate=separate),
                default=output.actual_value,
            )
            outp = cwl.CommandOutputParameter(
                id=sanitized_id,
                type=type_,
                outputBinding=cwl.CommandOutputBinding(glob=f"$(inputs.{arg.id})"),
            )
            return outp, arg

        return (
            cwl.CommandOutputParameter(
                id=sanitized_id, type=type_, outputBinding=cwl.CommandOutputBinding(glob=output.actual_value)
            ),
            None,
        )



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service cache delete uploaded chunks."""
import shutil

from renku.ui.service.config import CACHE_UPLOADS_PATH
from renku.ui.service.controllers.api.abstract import ServiceCtrl
from renku.ui.service.controllers.api.mixins import RenkuOperationMixin
from renku.ui.service.errors import IntermittentFileNotExistsError
from renku.ui.service.serializers.cache import FileChunksDeleteRequest, FileChunksDeleteResponseRPC
from renku.ui.service.views import result_response


class DeleteFileChunksCtrl(ServiceCtrl, RenkuOperationMixin):
    """Controller for deleting uploaded chunks."""

    REQUEST_SERIALIZER = FileChunksDeleteRequest()
    RESPONSE_SERIALIZER = FileChunksDeleteResponseRPC()

    def __init__(self, cache, user_data, request_data):
        """Construct list uploaded files controller."""
        self.ctx = DeleteFileChunksCtrl.REQUEST_SERIALIZER.load(request_data)
        super().__init__(cache, user_data, request_data)

    @property
    def context(self):
        """Controller operation context."""
        return self.ctx

    def delete_chunks(self):
        """Delete chunks from a chunked upload."""
        chunked_id = self.context["chunked_id"]

        user_cache_dir = CACHE_UPLOADS_PATH / self.user.user_id
        chunks_dir = user_cache_dir / chunked_id

        if not chunks_dir.exists():
            raise IntermittentFileNotExistsError(file_name=chunked_id)

        shutil.rmtree(chunks_dir)

        self.cache.invalidate_chunks(self.user, chunked_id)

        return f"Deleted chunks for {chunked_id}"

    def renku_op(self):
        """Renku operation for the controller."""
        # NOTE: We leave it empty since it does not execute renku operation.
        pass

    def to_response(self):
        """Execute controller flow and serialize to service response."""
        return result_response(DeleteFileChunksCtrl.RESPONSE_SERIALIZER, self.delete_chunks())



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Plugin hooks for renku workflow customization."""

from pathlib import Path
from typing import List, Optional, Protocol, Tuple

import pluggy

from renku.core import errors
from renku.domain_model.workflow.converters import IWorkflowConverter
from renku.domain_model.workflow.plan import Plan

hookspec = pluggy.HookspecMarker("renku")


@hookspec
def workflow_format() -> Tuple[IWorkflowConverter, List[str]]:  # type: ignore[empty-body]
    """Plugin Hook for ``workflow export`` call.

    Can be used to export renku workflows in different formats.

    Returns:
        Tuple[IWorkflowConverter,List[str]]: A tuple of the plugin itself and the output formats it supports.
            A plugin can support multiple formats.

    """
    pass


@hookspec(firstresult=True)
def workflow_convert(  # type: ignore[empty-body]
    workflow: Plan,
    basedir: Path,
    output: Optional[Path],
    output_format: Optional[str],
    resolve_paths: bool,
    nest_workflows: bool,
) -> str:
    """Plugin Hook for ``workflow export`` call.

    Can be used to export renku workflows in different formats.

    Args:
        workflow(Plan): A ``Plan`` object that describes the given workflow.
        basedir(Path): The base output directory.
        output(Optional[Path]): The output file, which will contain the workflow.
        output_format(Optional[str]): Output format supported by the given plugin.

    Returns:
        str: The string representation of the given Plan in the specific
            workflow format.

    """
    pass


def get_supported_formats() -> List[str]:
    """Returns the currently available workflow language format types.

    Returns:
        List of supported export formats.
    """
    from renku.core.plugin.pluginmanager import get_plugin_manager

    pm = get_plugin_manager()
    supported_formats = pm.hook.workflow_format()
    return [format for fs in supported_formats for format in fs[1]]


class WorkflowConverterProtocol(Protocol):
    """Typing protocol to specify type of the workflow converter hook."""

    def __call__(
        self,
        workflow: Plan,
        basedir: Path,
        output: Optional[Path] = None,
        output_format: Optional[str] = None,
        resolve_paths: Optional[bool] = None,
        nest_workflows: Optional[bool] = None,
    ) -> str:
        """Dummy method to let mypy know the type of the hook implementation."""
        raise NotImplementedError()


def workflow_converter(format: str) -> WorkflowConverterProtocol:
    """Returns a workflow converter function for a given format if available.

    Args:
        format(str): The format to convert to.

    Returns:
        The conversion plugin callable.
    """
    from renku.core.plugin.pluginmanager import get_plugin_manager

    pm = get_plugin_manager()
    supported_formats = pm.hook.workflow_format()
    export_plugins = list(map(lambda x: x[0], supported_formats))
    converter = list(map(lambda x: x[0], filter(lambda x: format in x[1], supported_formats)))
    if not any(converter):
        raise errors.ParameterError(f"The specified workflow exporter format '{format}' is not available.")
    elif len(converter) > 1:
        raise errors.ConfigurationError(
            f"The specified format '{format}' is supported by more than one export plugins!"
        )
    export_plugins.remove(converter[0])
    return pm.subset_hook_caller(name="workflow_convert", remove_plugins=export_plugins)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service cache file related models."""
import os
import shutil
from datetime import datetime

from walrus import BooleanField, DateTimeField, IntegerField, Model, TextField

from renku.ui.service.cache.base import BaseCache
from renku.ui.service.config import CACHE_UPLOADS_PATH


class File(Model):
    """User file object."""

    __database__ = BaseCache.model_db
    __namespace__ = BaseCache.namespace

    created_at = DateTimeField()

    file_id = TextField(primary_key=True, index=True)
    user_id = TextField(index=True)

    content_type = TextField()
    file_name = TextField()

    file_size = IntegerField()

    relative_path = TextField()
    is_archive = BooleanField()
    is_dir = BooleanField()
    unpack_archive = BooleanField()

    @property
    def abs_path(self):
        """Full path of cached file."""
        return CACHE_UPLOADS_PATH / self.user_id / self.relative_path

    @property
    def age(self):
        """Returns file's age in seconds."""
        # NOTE: `created_at` field is aligned to UTC timezone.
        age = int((datetime.utcnow() - self.created_at).total_seconds())
        return age

    def exists(self):
        """Ensure a file exists on file system."""
        if self.abs_path.exists():
            self.is_dir = self.abs_path.is_dir()
            return True

        return False

    def ttl_expired(self, ttl=None):
        """Check if file time to live has expired."""
        if not self.created_at:
            # If record does not contain created_at,
            # it means its an old record, and
            # we should mark it for deletion.
            return True

        ttl = ttl or int(os.getenv("RENKU_SVC_CLEANUP_TTL_FILES", 1800))
        return self.age >= ttl

    def purge(self):
        """Removes file from file system and cache."""
        if self.abs_path.is_file():
            self.abs_path.unlink()

        if self.abs_path.is_dir():
            shutil.rmtree(str(self.abs_path))

        self.delete()

    def is_locked(self, jobs):
        """Check if file locked by given jobs."""
        return bool(next((job for job in jobs if self.file_id in job.locked), False))


class FileChunk(Model):
    """User file chunk object."""

    __database__ = BaseCache.model_db
    __namespace__ = BaseCache.namespace

    created_at = DateTimeField()

    chunk_file_id = TextField(primary_key=True, index=True)
    user_id = TextField(index=True)
    chunked_id = TextField(index=True)

    file_name = TextField()
    relative_path = TextField()

    @property
    def abs_path(self):
        """Full path of cached file."""
        return CACHE_UPLOADS_PATH / self.user_id / self.chunked_id / self.relative_path

    @property
    def age(self):
        """Returns file's age in seconds."""
        # NOTE: `created_at` field is aligned to UTC timezone.
        age = int((datetime.utcnow() - self.created_at).total_seconds())
        return age

    def exists(self):
        """Ensure a file exists on file system."""
        if self.abs_path.exists():
            return True

        return False

    def ttl_expired(self, ttl=None):
        """Check if file time to live has expired."""
        if not self.created_at:
            # If record does not contain created_at,
            # it means its an old record, and
            # we should mark it for deletion.
            return True

        ttl = ttl or int(os.getenv("RENKU_SVC_CLEANUP_TTL_FILES", 1800))
        return self.age >= ttl

    def purge(self):
        """Removes file from file system and cache."""
        self.abs_path.unlink()
        self.delete()



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Recreate files created by the "run" command.

Commands and options
~~~~~~~~~~~~~~~~~~~~

.. rst-class:: cli-reference-commands

.. click:: renku.ui.cli.rerun:rerun
   :prog: renku rerun
   :nested: full

Examples
~~~~~~~~

.. image:: ../../_static/asciicasts/rerun.delay.gif
   :width: 850
   :alt: Rerun workflow

Assume you have run a step 2 that uses a stochastic algorithm, so each run
will be slightly different. The goal is to regenerate output ``C`` several
times to compare the output. In this situation it is not possible to simply
call :ref:`cli-update` since the input file ``A`` has not been modified
after the execution of step 2.

.. code-block:: text

    A-[step 1]-B-[step 2*]-C

Recreate a specific output file by running:

  .. code-block:: console

     $ renku rerun C

If you do not want step 1 to also be rerun, you can specify a starting point
using the ``--from`` parameter:

  .. code-block:: console

     $ renku rerun --from B C

Note that all other outputs of the executed workflow will be recreated as well.
If the output didn't change, it will be removed from git and re-added to ensure
that the re-execution is properly tracked.

In some cases it may be desirable to avoid updating the renku metadata
and to avoid committing this and any other change in the repository when the rerun
command is used. If this is the case then you can pass the ``--skip-metadata-update``
flag to ``renku rerun``.

.. cheatsheet::
   :group: Running
   :command: $ renku rerun <path>
   :description: Recreate the file(s) <path> by rerunning the commands that
                 created them.
   :target: rp
"""

import click
from lazy_object_proxy import Proxy

from renku.core import errors
from renku.ui.cli.utils.callback import ClickCallback
from renku.ui.cli.utils.plugins import available_workflow_providers


@click.command()
@click.option("--dry-run", "-n", is_flag=True, default=False, help="Show a preview of plans that will be executed.")
@click.option("--skip-metadata-update", is_flag=True, help="Do not update the metadata store for the execution.")
@click.option(
    "--from",
    "sources",
    type=click.Path(exists=True, dir_okay=False),
    multiple=True,
    help="Start an execution from this file.",
)
@click.argument("paths", type=click.Path(exists=True, dir_okay=True), nargs=-1, required=True)
@click.option(
    "provider",
    "-p",
    "--provider",
    default="toil",
    show_default=True,
    type=click.Choice(Proxy(available_workflow_providers), case_sensitive=False),
    help="The workflow engine to use.",
)
@click.option(
    "config", "-c", "--config", metavar="<config file>", help="YAML file containing configuration for the provider."
)
def rerun(dry_run, skip_metadata_update, sources, paths, provider, config):
    """Recreate files generated by a sequence of ``run`` commands."""
    from renku.command.format.activity import tabulate_activities
    from renku.command.rerun import rerun_command

    communicator = ClickCallback()

    try:
        result = (
            rerun_command(skip_metadata_update=skip_metadata_update)
            .with_communicator(communicator)
            .build()
            .execute(dry_run=dry_run, sources=sources, paths=paths, provider=provider, config=config)
        )
    except errors.NothingToExecuteError:
        exit(1)
    else:
        if dry_run:
            activities, modified_inputs = result.output
            click.echo(tabulate_activities(activities, modified_inputs))



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Migrate old metadata to new storage format."""

import os
import shutil
import traceback
import uuid
from collections import defaultdict
from hashlib import sha1
from itertools import chain
from pathlib import Path, PurePosixPath
from typing import TYPE_CHECKING, List, Optional, Union, cast
from urllib.parse import urlparse

import deal

from renku.command.command_builder import inject
from renku.core import errors
from renku.core.dataset.datasets_provenance import DatasetsProvenance
from renku.core.interface.activity_gateway import IActivityGateway
from renku.core.interface.database_gateway import IDatabaseGateway
from renku.core.interface.project_gateway import IProjectGateway
from renku.core.migration.models import v9 as old_schema
from renku.core.migration.models import v10 as new_schema
from renku.core.migration.models.migration import DatasetMigrationContext, MigrationContext, MigrationType
from renku.core.migration.utils import (
    OLD_DATASETS_PATH,
    OLD_METADATA_PATH,
    read_project_version_from_yaml,
    set_temporary_datasets_path,
    unset_temporary_datasets_path,
)
from renku.core.migration.utils.conversion import convert_dataset
from renku.core.util import communication
from renku.core.util.yaml import load_yaml
from renku.domain_model.constant import NON_EXISTING_ENTITY_CHECKSUM
from renku.domain_model.entity import Collection, Entity
from renku.domain_model.project_context import has_graph_files, project_context
from renku.domain_model.provenance.activity import Activity, Association, Generation, Usage
from renku.domain_model.provenance.agent import Person, SoftwareAgent
from renku.domain_model.provenance.parameter import ParameterValue
from renku.domain_model.workflow.parameter import (
    DIRECTORY_MIME_TYPE,
    CommandInput,
    CommandOutput,
    CommandParameter,
    MappedIOStream,
)
from renku.domain_model.workflow.plan import Plan

if TYPE_CHECKING:
    from renku.infrastructure.repository import Commit


def migrate(migration_context: MigrationContext):
    """Migration function."""
    repository = project_context.repository
    committed = _commit_previous_changes()
    # NOTE: Initialize submodules
    _ = repository.submodules

    migration_context.plan_cache = {}
    try:
        _generate_new_metadata(
            migration_context=migration_context,
            committed=committed,
            strict=migration_context.options.strict,
            migration_type=migration_context.options.type,
            preserve_identifiers=migration_context.options.preserve_identifiers,
        )
    finally:
        migration_context.plan_cache = None

    _remove_dataset_metadata_files()

    metadata_path = project_context.metadata_path.joinpath(OLD_METADATA_PATH)
    with open(metadata_path, "w") as f:
        f.write(
            "# Dummy file kept for backwards compatibility, does not contain actual version\n"
            "'http://schema.org/schemaVersion': '9'"
        )


def _commit_previous_changes():
    repository = project_context.repository
    if repository.is_dirty(untracked_files=False):
        project_path = project_context.metadata_path.joinpath(OLD_METADATA_PATH)
        project = old_schema.Project.from_yaml(project_path)
        project.version = "8"
        project.to_yaml(project_context.metadata_path.joinpath(project_path))

        repository.add(project_context.metadata_path)
        repository.commit(
            "renku migrate: committing structural changes" + project_context.transaction_id, no_verify=True
        )
        return True

    return False


@inject.autoparams()
def _maybe_migrate_project_to_database(project_gateway: IProjectGateway):
    """Migrate project to database if necessary."""
    metadata_path = project_context.metadata_path.joinpath(OLD_METADATA_PATH)

    if metadata_path.exists():
        old_project = cast(old_schema.Project, old_schema.Project.from_yaml(metadata_path))

        id_path = urlparse(old_project._id).path
        id_path = id_path.replace("/projects/", "")
        id_path = Path(id_path)
        namespace, name = str(id_path.parent), id_path.name
        id = new_schema.Project.generate_id(namespace=namespace, name=name)

        new_project = new_schema.Project(
            agent_version=old_project.agent_version,
            creator=_old_agent_to_new_agent(old_project.creator),
            date_created=old_project.created,
            id=id,
            name=old_project.name,
            version=old_project.version,
            template_id=old_project.template_id,
            template_metadata=old_project.template_metadata,
            template_ref=old_project.template_ref,
            template_source=old_project.template_source,
            template_version=old_project.template_version,
            immutable_template_files=old_project.immutable_template_files,
            automated_update=old_project.automated_update,
        )

        project_gateway.update_project(new_project)


def _remove_graph_files():
    """Remove all graph files."""
    # NOTE: These are required for projects that have new graph files
    try:
        (project_context.path / "provenance.json").unlink()
    except FileNotFoundError:
        pass
    try:
        (project_context.path / "dependency.json").unlink()
    except FileNotFoundError:
        pass
    try:
        shutil.rmtree(project_context.database_path)
    except FileNotFoundError:
        pass
    try:
        (project_context.path / "dataset.json").unlink()
    except FileNotFoundError:
        pass


@inject.autoparams("database_gateway", "activity_gateway")
def _generate_new_metadata(
    migration_context: MigrationContext,
    strict: bool,
    migration_type: MigrationType,
    database_gateway: IDatabaseGateway,
    activity_gateway: IActivityGateway,
    force=True,
    remove=True,
    committed=False,
    preserve_identifiers=False,
):
    """Generate graph and dataset provenance metadata."""
    repository = project_context.repository

    if force:
        _remove_graph_files()
    elif has_graph_files():
        raise errors.OperationError("Graph metadata exists.")

    database_gateway.initialize()

    _maybe_migrate_project_to_database()

    datasets_provenance = DatasetsProvenance()

    commits = list(
        repository.iterate_commits(
            f"{project_context.metadata_path}/workflow/*.yaml", ".renku/datasets/*/*.yml", reverse=True
        )
    )
    n_commits = len(commits)

    for n, commit in enumerate(commits, start=1):
        communication.echo(f"Processing commits {n}/{n_commits} {commit.hexsha}", end="\r")

        # NOTE: Treat the last commit differently if it was done by this migration
        is_last_commit = committed and n == n_commits

        try:
            # NOTE: Don't migrate workflows for dataset-only migrations
            if MigrationType.WORKFLOWS in migration_type:
                _process_workflows(
                    migration_context=migration_context, activity_gateway=activity_gateway, commit=commit, remove=remove
                )
            _process_datasets(
                migration_context=migration_context,
                commit=commit,
                datasets_provenance=datasets_provenance,
                is_last_commit=is_last_commit,
                preserve_identifiers=preserve_identifiers,
            )
        except errors.MigrationError:
            if strict:
                raise
            communication.echo("")
            communication.warn(f"Cannot process commit '{commit.hexsha}' - Migration failed: {traceback.format_exc()}")
        except Exception:
            if strict:
                raise
            communication.echo("")
            communication.warn(f"Cannot process commit '{commit.hexsha}' - Exception: {traceback.format_exc()}")

        # NOTE: Commit changes after each step
        database_gateway.commit()

    database_gateway.commit()


def _convert_run_to_plan(run: old_schema.Run, migration_context: MigrationContext, project_id) -> Plan:
    """Create a Plan from a Run."""
    assert not run.subprocesses, f"Cannot create a Plan from a Run with subprocesses: {run._id}"

    def extract_run_uuid(run_id: str) -> str:
        # https://localhost/runs/723fd784-9347-4081-84de-a6dbb067545b/
        return run_id.rstrip("/").rsplit("/", maxsplit=1)[-1]

    run_uuid = extract_run_uuid(run._id)

    if run_uuid in migration_context.plan_cache:
        return migration_context.plan_cache[run_uuid]

    plan_id = Plan.generate_id(uuid=run_uuid)

    def get_mime_type(entity: Union[old_schema.Entity, old_schema.Collection]) -> List[str]:
        return [DIRECTORY_MIME_TYPE] if isinstance(entity, old_schema.Collection) else ["application/octet-stream"]

    def convert_argument(argument: old_schema.CommandArgument) -> CommandParameter:
        """Convert an old CommandArgument to a new CommandParameter."""
        assert isinstance(argument, old_schema.CommandArgument)

        return CommandParameter(
            default_value=argument.value,
            description=argument.description,
            id=CommandParameter.generate_id(plan_id=plan_id, postfix=PurePosixPath(argument._id).name),
            name=argument.name,
            position=argument.position,
            prefix=argument.prefix,
            postfix=PurePosixPath(argument._id).name,
        )

    def convert_input(input: old_schema.CommandInput) -> CommandInput:
        """Convert an old CommandInput to a new CommandInput."""
        assert isinstance(input, old_schema.CommandInput)

        mapped_to = input.mapped_to
        if mapped_to:
            mapped_to = MappedIOStream(stream_type=mapped_to.stream_type)

        return CommandInput(
            default_value=input.consumes.path,
            description=input.description,
            id=CommandInput.generate_id(plan_id=plan_id, postfix=PurePosixPath(input._id).name),
            mapped_to=mapped_to,
            name=input.name,
            position=input.position,
            prefix=input.prefix,
            encoding_format=get_mime_type(input.consumes),
            postfix=PurePosixPath(input._id).name,
        )

    def convert_output(output: old_schema.CommandOutput) -> CommandOutput:
        """Convert an old CommandOutput to a new CommandOutput."""
        assert isinstance(output, old_schema.CommandOutput)

        mapped_to = output.mapped_to
        if mapped_to:
            mapped_to = MappedIOStream(stream_type=mapped_to.stream_type)

        return CommandOutput(
            create_folder=output.create_folder,
            default_value=output.produces.path,
            description=output.description,
            id=CommandOutput.generate_id(plan_id=plan_id, postfix=PurePosixPath(output._id).name),
            mapped_to=mapped_to,
            name=output.name,
            position=output.position,
            prefix=output.prefix,
            encoding_format=get_mime_type(output.produces),
            postfix=PurePosixPath(output._id).name,
        )

    plan = Plan(
        command=run.command,
        description=run.description,
        id=plan_id,
        inputs=[convert_input(i) for i in run.inputs],
        keywords=run.keywords,
        name=run.name,
        outputs=[convert_output(o) for o in run.outputs],
        parameters=[convert_argument(a) for a in run.arguments],
        project_id=project_id,
        success_codes=run.successcodes,
    )

    migration_context.plan_cache[run_uuid] = plan

    return plan


def _get_process_runs(workflow_run: old_schema.WorkflowRun) -> List[old_schema.ProcessRun]:
    # NOTE: Use Plan subprocesses to get activities because it is guaranteed to have correct order
    assert (
        workflow_run.association.plan.subprocesses
    ), "Can only get subprocesses if the associated plan has subprocesses"
    sorted_ids = [s.process._id for s in workflow_run.association.plan.subprocesses]
    activities = []
    # NOTE: it's possible to have subprocesses with similar ids but it does not matter since they have the same
    # plan
    # TODO: Remove these redundant subprocesses

    for id_ in sorted_ids:
        for s in workflow_run.subprocesses.values():
            if s.association.plan._id == id_:
                activities.append(s)
                break
    assert len(activities) == len(workflow_run.subprocesses)
    return activities


def _process_workflows(
    migration_context: MigrationContext, activity_gateway: IActivityGateway, commit: "Commit", remove: bool
):
    try:
        deal.disable(warn=False)
        for file in commit.get_changes(f"{project_context.metadata_path}/workflow/*.yaml"):
            if file.deleted:
                continue

            path: str = file.b_path

            if not path.startswith(".renku/workflow") or not path.endswith(".yaml"):
                continue

            if not (project_context.path / path).exists():
                communication.warn(f"Workflow file does not exists: '{path}'")
                continue

            workflow = old_schema.Activity.from_yaml(path=path)

            if isinstance(workflow, old_schema.WorkflowRun):
                activities = _get_process_runs(workflow)
            else:
                activities = [workflow]

            for old_activity in activities:
                new_activities = _process_run_to_new_activity(
                    migration_context=migration_context, process_run=old_activity
                )
                for new_activity in new_activities:
                    activity_gateway.add(new_activity)

            if remove:
                try:
                    os.remove(file.b_path)
                except FileNotFoundError:
                    pass
    finally:
        deal.enable()


def _process_run_to_new_activity(
    migration_context: MigrationContext, process_run: old_schema.ProcessRun
) -> List[Activity]:
    """Convert a ProcessRun to a new Activity."""

    def generate_activity_id(process_run_id: str, suffix: str = None) -> str:
        # https://localhost/activities/commit/b092dca8866c940dd4ff119b7d76d163a1be6a65
        # https://localhost/activities/commit/b092dca8866c940dd4ff119b7d76d163a1be6a65/steps/step_1
        parts = process_run_id.rstrip("/").rsplit("/", maxsplit=3)

        if parts[-2] == "steps":
            parts = parts[-3:]
        else:
            parts = [parts[-1]]

        original_id = "/".join(parts)

        if suffix:
            original_id += suffix

        return Activity.generate_id(sha1(original_id.encode("utf-8")).hexdigest())

    assert not isinstance(process_run, old_schema.WorkflowRun)

    project_id = project_context.project.id

    run = process_run.association.plan

    if run.subprocesses:
        runs = [s.process for s in run.subprocesses]
    else:
        runs = [run]

    activities = []
    for i, run in enumerate(runs):
        activity_id = generate_activity_id(process_run._id, suffix=str(i) if i else None)
        plan = _convert_run_to_plan(run=run, migration_context=migration_context, project_id=project_id)

        agents = [_old_agent_to_new_agent(a) for a in process_run.agents or []]
        association_agent = _old_agent_to_new_agent(process_run.association.agent)
        association = Association(agent=association_agent, id=Association.generate_id(activity_id), plan=plan)

        # NOTE: The same entity can have the same id during different times in its lifetime (e.g. different commit_sha,
        # but the same content). When it gets flattened, some fields will have multiple values which will cause an error
        # during deserialization. Make sure that no such Entity attributes exists (store those information in the
        # Generation object).

        invalidations = [_convert_invalidated_entity(e) for e in (process_run.invalidated or [])]

        generations = []

        for potential_generation in chain(process_run.generated or [], process_run.qualified_usage or []):
            if all(output.produces.path != potential_generation.entity.path for output in run.outputs):
                continue

            generations.append(_convert_generation(potential_generation, activity_id))

        usages = []
        for potential_usage in process_run.qualified_usage or []:
            if all(input.consumes.path != potential_usage.entity.path for input in run.inputs):
                continue

            usages.append(_convert_usage(potential_usage, activity_id))

        parameters = _create_parameters(activity_id=activity_id, plan=plan, usages=usages, generations=generations)

        activities.append(
            Activity(
                agents=agents,
                annotations=process_run.annotations,
                association=association,
                ended_at_time=process_run.ended_at_time,
                generations=generations,
                id=activity_id,
                invalidations=invalidations,
                parameters=parameters,
                project_id=project_id,
                started_at_time=process_run.started_at_time,
                usages=usages,
            )
        )

    return activities


def _convert_usage(usage: old_schema.Usage, activity_id: str) -> Usage:
    """Convert an old qualified Usage to a new one."""
    commit_sha = _extract_commit_sha(entity_id=usage.entity._id)
    entity = _convert_used_entity(usage.entity, commit_sha, activity_id)
    assert entity, f"Top entity was not found for Usage: {usage._id}, {usage.entity.path}"

    return Usage(id=Usage.generate_id(activity_id), entity=entity)


def _convert_generation(generation: old_schema.Generation, activity_id: str) -> Generation:
    """Convert an old Generation to a new one."""
    commit_sha = _extract_commit_sha(entity_id=generation.entity._id)
    entity = _convert_generated_entity(generation.entity, commit_sha, activity_id)
    assert entity, f"Root entity was not found for Generation: {generation._id}"

    return Generation(id=Generation.generate_id(activity_id), entity=entity)


def _convert_used_entity(entity: old_schema.Entity, revision: str, activity_id: str) -> Entity:
    """Convert an old Entity to one with proper metadata.

    For Collections, add members that are modified in the same commit or before the revision.
    """
    assert isinstance(entity, old_schema.Entity)

    repository = project_context.repository

    checksum = repository.get_object_hash(revision=revision, path=entity.path)
    if not checksum:
        _EntityWarningCache.warn(path=entity.path, revision=revision)
        checksum = NON_EXISTING_ENTITY_CHECKSUM

    if isinstance(entity, old_schema.Collection):
        members = []
        for child in entity.members:
            new_child = _convert_used_entity(child, revision, activity_id)
            if not new_child:
                continue
            members.append(new_child)

        new_entity = Collection(checksum=checksum, path=entity.path, members=members)
    else:
        new_entity = Entity(checksum=checksum, path=entity.path)

    assert new_entity.__class__.__name__ == entity.__class__.__name__

    return new_entity


def _convert_generated_entity(entity: old_schema.Entity, revision: str, activity_id: str) -> Optional[Entity]:
    """Convert an Entity to one with proper metadata.

    For Collections, add members that are modified in the same commit as revision.
    """
    assert isinstance(entity, old_schema.Entity)

    repository = project_context.repository

    try:
        entity_commit = repository.get_previous_commit(path=entity.path, revision=revision, submodule=True)
    except errors.GitCommitNotFoundError:
        return None

    if entity_commit.hexsha != revision:
        return None

    checksum = repository.get_object_hash(revision=revision, path=entity.path)
    if not checksum:
        _EntityWarningCache.warn(path=entity.path, revision=revision)
        checksum = NON_EXISTING_ENTITY_CHECKSUM

    if isinstance(entity, old_schema.Collection):
        members = []
        for child in entity.members:
            new_child = _convert_generated_entity(child, revision, activity_id)
            if not new_child:
                continue
            members.append(new_child)

        new_entity = Collection(checksum=checksum, path=entity.path, members=members)
    else:
        new_entity = Entity(checksum=checksum, path=entity.path)

    assert new_entity.__class__.__name__ == entity.__class__.__name__

    return new_entity


def _convert_invalidated_entity(entity: old_schema.Entity) -> Optional[Entity]:
    """Convert an Entity to one with proper metadata."""
    assert isinstance(entity, old_schema.Entity)
    assert not isinstance(entity, old_schema.Collection), f"Collection passed as invalidated: {entity._id}"

    repository = project_context.repository

    commit_sha = _extract_commit_sha(entity_id=entity._id)
    commit = repository.get_previous_commit(revision=commit_sha, path=entity.path, submodule=True)
    revision = commit.hexsha
    checksum = repository.get_object_hash(revision=revision, path=entity.path)
    if not checksum:
        # Entity was deleted at revision; get the one before it to have object_id
        checksum = repository.get_object_hash(revision=f"{revision}~", path=entity.path)
        if not checksum:
            _EntityWarningCache.warn(path=entity.path, revision=revision)
            checksum = NON_EXISTING_ENTITY_CHECKSUM

    new_entity = Entity(checksum=checksum, path=entity.path)

    assert new_entity.__class__.__name__ == entity.__class__.__name__

    return new_entity


def _extract_commit_sha(entity_id: str) -> str:
    # NOTE: extracts commit sha from ids like /blob/a3bf8a165dd56da078b96f2ca2ff22f14a3bdd57/input
    path = urlparse(entity_id).path
    assert path.startswith("/blob/"), f"Invalid entity identifier: {entity_id}"

    commit_sha = path[len("/blob/") :].split("/", 1)[0]
    assert len(commit_sha) == 40, f"Entity does not have valid commit SHA: {entity_id}"

    return commit_sha


def _create_parameters(activity_id, plan: Plan, usages: List[Usage], generations: List[Generation]):
    parameters = []

    inputs = {i.default_value: i for i in plan.inputs}

    for usage in usages:
        input = inputs.pop(usage.entity.path, None)
        assert input is not None, f"Cannot find usage path '{usage.entity.path}' in plan {plan.id}"
        id = ParameterValue.generate_id(activity_id)
        parameters.append(ParameterValue(id=id, parameter_id=input.id, value=usage.entity.path))

    outputs = {o.default_value: o for o in plan.outputs}

    for generation in generations:
        output = outputs.pop(generation.entity.path, None)
        assert output is not None, f"Cannot find generation path '{generation.entity.path}' in plan {plan.id}"
        id = ParameterValue.generate_id(activity_id)
        parameters.append(ParameterValue(id=id, parameter_id=output.id, value=generation.entity.path))

    for parameter in plan.parameters:
        id = ParameterValue.generate_id(activity_id)
        parameters.append(ParameterValue(id=id, parameter_id=parameter.id, value=parameter.default_value))

    return parameters


def _old_agent_to_new_agent(
    agent: Optional[Union[old_schema.Person, old_schema.SoftwareAgent]]
) -> Optional[Union[Person, SoftwareAgent]]:
    """Create an instance from Person/SoftwareAgent."""
    if isinstance(agent, old_schema.SoftwareAgent):
        return SoftwareAgent(id=agent.id, name=agent.label)

    assert not agent or isinstance(agent, old_schema.Person), f"Invalid type {type(agent)}"
    return Person(
        affiliation=agent.affiliation,
        alternate_name=agent.alternate_name,
        email=agent.email,
        id=None,
        name=agent.name,
    )


def _process_datasets(
    migration_context: MigrationContext,
    commit: "Commit",
    datasets_provenance: DatasetsProvenance,
    is_last_commit,
    preserve_identifiers,
):
    changes = commit.get_changes(".renku/datasets/*/*.yml")
    changed_paths = [c.b_path for c in changes if not c.deleted]
    paths = [p for p in changed_paths if len(Path(p).parents) == 4]  # Exclude files that are not in the right place
    deleted_paths = [c.a_path for c in changes if c.deleted]
    deleted_paths = [p for p in deleted_paths if len(Path(p).parents) == 4]

    datasets, deleted_datasets = fetch_datasets(
        migration_context=migration_context, revision=commit.hexsha, paths=paths, deleted_paths=deleted_paths
    )

    revision = commit.hexsha
    date = commit.authored_datetime

    for dataset in datasets:
        dataset, tags = convert_dataset(dataset=dataset, revision=revision)

        datasets_provenance.update_during_migration(
            dataset,
            commit_sha=revision,
            date=date,
            tags=tags,
            replace=True if is_last_commit else False,
            preserve_identifiers=preserve_identifiers,
        )

    for dataset in deleted_datasets:
        dataset, _ = convert_dataset(dataset=dataset, revision=revision)
        datasets_provenance.update_during_migration(
            dataset, commit_sha=revision, date=date, remove=True, preserve_identifiers=preserve_identifiers
        )


def fetch_datasets(
    migration_context: MigrationContext,
    revision: str,
    paths: List[str],
    deleted_paths: List[str],
):
    """Fetch a dataset from a given revision."""
    from renku.core.migration.models.v9 import Dataset

    repository = project_context.repository

    datasets_path = repository.path / ".renku" / "tmp" / OLD_DATASETS_PATH
    shutil.rmtree(datasets_path, ignore_errors=True)
    datasets_path.mkdir(parents=True, exist_ok=True)

    def read_project_version():
        """Read project version at revision."""
        try:
            project_file_content = repository.get_content(path=".renku/metadata.yml", revision=revision)
        except errors.GitCommandError:  # Project metadata file does not exist
            return 1

        try:
            yaml_data = load_yaml(project_file_content)
            return int(read_project_version_from_yaml(yaml_data))
        except ValueError:
            return 1

    def get_dataset_identifier(path):
        identifier = Path(path).parent.name
        try:
            identifier = str(uuid.UUID(identifier))
        except ValueError:
            identifier = identifier.rsplit("%2F", maxsplit=1)[-1]
            try:
                identifier = str(uuid.UUID(identifier))
            except ValueError:
                return
            else:
                return identifier
        else:
            return identifier

    def copy_and_migrate_datasets():
        existing = []
        deleted = []

        for path in chain(paths, deleted_paths):
            rev = revision
            if path in deleted_paths:
                rev = repository.get_previous_commit(path, revision=f"{revision}~", submodule=True)
            identifier = get_dataset_identifier(path)
            if not identifier:
                continue
            new_path = datasets_path / identifier / "metadata.yml"
            new_path.parent.mkdir(parents=True, exist_ok=True)
            content = repository.get_content(path=path, revision=str(rev))
            new_path.write_text(content)
            if path in deleted_paths:
                deleted.append(new_path)
            else:
                existing.append(new_path)

        migration_context.dataset_migration_context = DatasetMigrationContext(repository=repository, revision=revision)

        try:
            from renku.core.migration.migrate import migrate_project

            project_version = read_project_version()
            set_temporary_datasets_path(datasets_path)
            communication.disable()

            with project_context.with_path(repository.path):
                migrate_project(
                    project_version=project_version,
                    skip_template_update=True,
                    skip_docker_update=True,
                    max_version=8,
                    migration_type=MigrationType.DATASETS,
                )
        finally:
            communication.enable()
            unset_temporary_datasets_path()
            migration_context.dataset_migration_context = None

        return existing, deleted

    paths, deleted_paths = copy_and_migrate_datasets()

    datasets = []
    for metadata_path in paths:
        dataset = Dataset.from_yaml(metadata_path)
        # NOTE: Fixing dataset path after migration
        initial_identifier = Path(dataset.path).name
        dataset.path = f".renku/datasets/{initial_identifier}"
        datasets.append(dataset)

    deleted_datasets = []
    for metadata_path in deleted_paths:
        dataset = Dataset.from_yaml(metadata_path)
        # NOTE: Fixing dataset path after migration
        initial_identifier = Path(dataset.path).name
        dataset.path = f".renku/datasets/{initial_identifier}"
        deleted_datasets.append(dataset)

    return datasets, deleted_datasets


def _remove_dataset_metadata_files():
    """Remove old dataset metadata."""
    try:
        shutil.rmtree(os.path.join(project_context.metadata_path, OLD_DATASETS_PATH))
    except FileNotFoundError:
        pass
    try:
        shutil.rmtree(os.path.join(project_context.metadata_path, "refs", OLD_DATASETS_PATH))
    except FileNotFoundError:
        pass


class _EntityWarningCache:
    """Cache warning messages to avoid re-prints."""

    cache = defaultdict(set)

    @staticmethod
    def warn(path, revision):
        """Print a warning if it's not already printed for the path/revision."""
        path = str(path)
        revision = str(revision)

        already_warned = _EntityWarningCache.cache[path]
        if revision in already_warned:
            return

        already_warned.add(revision)
        communication.warn(f"Entity '{path}' not found at '{revision}'")



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""toil based provider."""

import functools
import itertools
import os
import shutil
import uuid
from abc import abstractmethod
from pathlib import Path
from subprocess import call
from typing import Any, Callable, Dict, List, Union, cast

import networkx as nx
from toil.common import Toil
from toil.fileStores import FileID
from toil.fileStores.abstractFileStore import AbstractFileStore
from toil.job import Job, Promise
from toil.leader import FailedJobsException
from toil.lib.docker import apiDockerCall

from renku.command.util import progressbar
from renku.core import errors
from renku.core.constant import RENKU_HOME, RENKU_TMP
from renku.core.errors import WorkflowExecuteError
from renku.core.plugin import hookimpl
from renku.core.plugin.provider import RENKU_ENV_PREFIX
from renku.domain_model.workflow.parameter import CommandParameterBase
from renku.domain_model.workflow.plan import Plan
from renku.domain_model.workflow.provider import IWorkflowProvider


class AbstractToilJob(Job):
    """Toil job implementation for a renku ``Plan``."""

    def __init__(self, workflow: Plan, *args, **kwargs):
        super().__init__(unitName=workflow.name, displayName=workflow.name, *args, **kwargs)
        self.workflow: Plan = workflow
        self._input_files: Dict[str, FileID] = {}
        self._parents_promise: List[Promise] = []
        self._environment = os.environ.copy()

    @abstractmethod
    def _execute(self, command_line: List[str], mapped_std: Dict[str, str]) -> int:
        """Executes a given command line."""
        raise NotImplementedError

    def set_input_files(self, input_files: Dict[str, Any]):
        """Set the location of inputs that are available in the original project."""
        self._input_files = input_files

    def add_input_promise(self, promise: Dict[str, FileID]):
        """Adds a given job's output promise.

        The promise of a job contains its output locations in the global storage.
        """
        self._parents_promise.append(promise)

    def run(self, storage):
        """Executing of a renku ``Plan``."""
        mapped_std = dict()
        parent_inputs = dict()
        for p in self._parents_promise:
            parent_inputs.update(p)

        def _read_input(input: str, file_metadata):
            input_path = Path(input)

            if isinstance(file_metadata, dict):
                input_path.mkdir(parents=True, exist_ok=True)
                for path, file_id in file_metadata.items():
                    _read_input(path, file_id)
            elif not input_path.exists():
                if len(input_path.parts) > 1:
                    input_path.parent.mkdir(parents=True, exist_ok=True)
                storage.readGlobalFile(file_metadata, userPath=input)

        for i in self.workflow.inputs:
            file_metadata = (
                parent_inputs[i.actual_value] if i.actual_value in parent_inputs else self._input_files[i.actual_value]
            )
            _read_input(i.actual_value, file_metadata)

            self._environment[f"{RENKU_ENV_PREFIX}{i.name}"] = str(i.actual_value)

            if i.mapped_to:
                mapped_std[i.mapped_to.stream_type] = i.actual_value

        for o in self.workflow.outputs:
            self._environment[f"{RENKU_ENV_PREFIX}{o.name}"] = str(o.actual_value)
            output_path = Path(o.actual_value)
            if len(output_path.parts) > 1:
                output_path.parent.mkdir(parents=True, exist_ok=True)

            if o.mapped_to:
                mapped_std[o.mapped_to.stream_type] = o.actual_value

        for p in self.workflow.parameters:
            self._environment[f"{RENKU_ENV_PREFIX}{p.name}"] = str(p.actual_value)

        # construct cmd
        cmd = []

        if self.workflow.command:
            cmd.extend(self.workflow.command.split(" "))

        arguments = itertools.chain(self.workflow.inputs, self.workflow.outputs, self.workflow.parameters)

        arguments_filtered = filter(lambda x: x.position is not None and not getattr(x, "mapped_to", None), arguments)
        arguments_sorted = sorted(arguments_filtered, key=lambda x: cast(int, x.position))

        for a in arguments_sorted:
            v = str(a.actual_value) if not isinstance(a.actual_value, str) else a.actual_value
            if a.prefix:
                if a.prefix.endswith(" "):
                    cmd.append(a.prefix[:-1])
                else:
                    v = f"{a.prefix}{v}"
            cmd.append(v)

        return_code = self._execute(cmd, mapped_std)
        if return_code not in (self.workflow.success_codes or {0}):
            raise errors.InvalidSuccessCode(return_code, success_codes=self.workflow.success_codes)

        return _upload_files(
            storage.writeGlobalFile, cast(List[CommandParameterBase], self.workflow.outputs), Path.cwd()
        )


class SubprocessToilJob(AbstractToilJob):
    """A toil job that uses subprocess.call to execute a renku ``Plan``."""

    def __init__(self, workflow: Plan, *args, **kwargs):
        super().__init__(workflow, *args, **kwargs)

    def _execute(self, command_line: List[str], mapped_std: Dict[str, str]) -> int:
        """Executes a given command line."""
        return call(
            command_line,
            cwd=os.getcwd(),
            **{  # type: ignore
                key: open(value, mode="r" if key == "stdin" else "w")
                for key, value in mapped_std.items()  # type: ignore
            },
            env=self._environment,
        )


class DockerToilJob(AbstractToilJob):
    """A toil job that uses apiDockerCall to execute a renku ``Plan``."""

    def __init__(self, workflow: Plan, docker_config: Dict[str, Any], *args, **kwargs):
        super().__init__(workflow, *args, **kwargs)
        self._docker_config: Dict[str, Any] = docker_config

    def _execute(self, command_line: List[str], mapped_std: Dict[str, str]) -> int:
        """Executes a given command line."""
        # NOTE: Disable detached mode to block for ``apiDockerCall`` to finish
        self._docker_config.pop("detach", None)

        parameters: Union[List[str], List[List[str]]] = command_line

        stderr = mapped_std.get("stderr")
        stdin = mapped_std.get("stdin")
        if stderr or stdin:
            raise errors.OperationError("Cannot run workflows that have stdin or stderr redirection with Docker")
        stdout = mapped_std.get("stdout")
        if stdout:
            parameters = [command_line, ["tee", stdout]]

        volumes = self._docker_config.pop("volumes", {})

        if "working_dir" not in self._docker_config:
            working_dir = "/renku"
            volumes[os.getcwd()] = {"bind": working_dir, "mode": "rw"}
        else:
            working_dir = self._docker_config.pop("working_dir")

        # NOTE: We cannot get the exit code back from the docker container. The Docker API checks for the exit code and
        # raises an exception if it's not 0, so, Plan.success_codes is ignored when running with Docker.
        apiDockerCall(
            self,
            parameters=parameters,
            detach=False,
            environment=self._environment,
            volumes=volumes,
            working_dir=working_dir,
            **self._docker_config,
        )

        return 0


def _store_location(
    import_function: Callable[[str], FileID], basedir: Path, location: Path
) -> Union[FileID, Dict[str, Any]]:
    if location.is_dir():
        directory_content = dict()
        for f in location.rglob("*"):
            directory_content[str(f.relative_to(basedir))] = _store_location(import_function, basedir, f)
        return directory_content
    else:
        return import_function(str(location))


def _upload_files(
    import_function: Callable[[str], FileID], params: List[CommandParameterBase], basedir: Path
) -> Dict[str, FileID]:
    file_locations = dict()
    for p in params:
        location = basedir / p.actual_value
        if not location.exists():
            continue

        file_locations[p.actual_value] = _store_location(import_function, basedir, location)

    return file_locations


def import_file_wrapper(storage: AbstractFileStore, file_path: str) -> FileID:
    """Wrap importFile accept file:// URIs."""
    file_uri = file_path if ":/" in file_path else f"file://{file_path}"
    return storage.importFile(file_uri)


def process_children(
    parent: AbstractToilJob,
    dag: nx.DiGraph,
    jobs: Dict[int, AbstractToilJob],
    basedir: Path,
    storage: AbstractFileStore,
):
    """Recursively process children of a workflow."""
    outputs = list()
    import_function = functools.partial(import_file_wrapper, storage)
    for child in nx.neighbors(dag, parent.workflow):
        child_job = jobs[id(child)]
        file_metadata = _upload_files(import_function, child.inputs, basedir)
        child_job.set_input_files(file_metadata)
        child_job.add_input_promise(parent.rv())
        outputs.append(parent.addFollowOn(child_job).rv())
        outputs += process_children(child_job, dag, jobs, basedir, storage)
    return outputs


def initialize_jobs(job, basedir, dag, docker_config):
    """Creates the Toil execution plan for the given workflow DAG."""
    job.fileStore.logToMaster("executing renku DAG")
    outputs = list()
    if docker_config:
        job.fileStore.logToMaster("executing with Docker")
        jobs = {id(n): DockerToilJob(n, docker_config) for n in dag.nodes}
    else:
        jobs = {id(n): SubprocessToilJob(n) for n in dag.nodes}
    import_function = functools.partial(import_file_wrapper, job.fileStore)
    children = next(nx.topological_generations(dag))
    for workflow in children:
        child_job = jobs[id(workflow)]
        file_metadata = _upload_files(import_function, workflow.inputs, basedir)
        child_job.set_input_files(file_metadata)
        outputs.append(job.addChild(child_job).rv())
        outputs += process_children(child_job, dag, jobs, basedir, job.fileStore)

    return outputs


class ToilProvider(IWorkflowProvider):
    """A workflow executor provider using toil."""

    @hookimpl
    def workflow_provider(self):
        """Workflow provider name."""
        return self, "toil"

    @hookimpl
    def workflow_execute(self, dag: nx.DiGraph, basedir: Path, config: Dict[str, Any]):
        """Executes a given workflow DAG using Toil."""
        tmpdir = (basedir / RENKU_HOME / RENKU_TMP).resolve()
        if not tmpdir.exists():
            tmpdir.mkdir()
        options = Job.Runner.getDefaultOptions(str(tmpdir / uuid.uuid4().hex))
        options.logLevel = "ERROR"
        options.clean = "always"

        config = config or {}
        docker_config = config.pop("docker", {})
        if docker_config and "image" not in docker_config:
            raise errors.ConfigurationError("Docker configuration must provide an 'image' property")

        if config:
            for k, v in config.items():
                setattr(options, k, v)

        outputs = list()
        try:
            with Toil(options) as toil:
                root_job = Job.wrapJobFn(initialize_jobs, basedir, dag, docker_config)
                job_outputs = toil.start(root_job)

                num_outputs = sum(map(lambda x: len(x.values()), job_outputs))
                with progressbar(length=num_outputs, label="Moving outputs") as bar:
                    for collection in job_outputs:
                        for name, fid in collection.items():
                            if isinstance(fid, dict):
                                directory = basedir / name
                                if directory.exists():
                                    shutil.rmtree(str(directory))
                                directory.mkdir()

                                for path, file_id in fid.items():
                                    destination = (basedir / path).absolute()
                                    destination.parent.mkdir(parents=True, exist_ok=True)
                                    toil.exportFile(file_id, str(destination))
                                    outputs.append(path)
                            else:
                                destination = (basedir / name).absolute()
                                destination.parent.mkdir(parents=True, exist_ok=True)
                                toil.exportFile(fid, str(destination))
                                outputs.append(name)
                            bar.update(1)
        except FailedJobsException as e:
            raise WorkflowExecuteError(e.msg)

        return outputs



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Serializers for activities."""

from datetime import datetime
from typing import List, NamedTuple, Set

from renku.command.format.tabulate import tabulate
from renku.core.util.os import are_paths_related
from renku.domain_model.provenance.activity import Activity


def tabulate_activities(activities: List[Activity], modified_inputs: Set[str]):
    """Return some info about the activities in a tabular form.

    Args:
        activities(List[Activity]): Activities to convert.
        modified_inputs(Set[str]): Set of modified inputs for activities.

    Returns:
        String of activities in tabular representation.
    """
    collection = []
    fields = "plan,execution_date,modified_inputs,outputs,command"

    class ActivityDisplay(NamedTuple):
        plan: str
        execution_date: datetime
        modified_inputs: str
        outputs: str
        command: str

    for activity in activities:
        modified_usages = {
            u.entity.path for u in activity.usages if any(are_paths_related(u.entity.path, m) for m in modified_inputs)
        }
        generations = {g.entity.path for g in activity.generations}
        modified_inputs |= generations
        plan = activity.association.plan.name

        collection.append(
            ActivityDisplay(
                plan,
                activity.ended_at_time,
                ", ".join(sorted(modified_usages)),
                ", ".join(sorted(generations)),
                " ".join(activity.plan_with_values.to_argv(with_streams=True)),
            )
        )

    return tabulate(collection=collection, columns=fields, columns_mapping=ACTIVITY_DISPLAY_COLUMNS, sort=False)


ACTIVITY_DISPLAY_COLUMNS = {
    "plan": ("plan", None),
    "execution_date": ("execution_date", "date executed"),
    "modified_inputs": ("modified_inputs", "modified inputs"),
    "outputs": ("outputs", None),
    "command": ("command", None),
}



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service template read manifest controller."""
import base64
from io import BytesIO

from marshmallow import EXCLUDE

from renku.core.template.template import fetch_templates_source
from renku.core.util.os import get_safe_relative_path
from renku.ui.service.controllers.api.abstract import ServiceCtrl
from renku.ui.service.controllers.api.mixins import RenkuOperationMixin
from renku.ui.service.serializers.templates import ManifestTemplatesRequest, ManifestTemplatesResponseRPC
from renku.ui.service.views import result_response

MAX_ICON_SIZE = (256, 256)


class TemplatesReadManifestCtrl(ServiceCtrl, RenkuOperationMixin):
    """Template read manifest controller."""

    REQUEST_SERIALIZER = ManifestTemplatesRequest()
    RESPONSE_SERIALIZER = ManifestTemplatesResponseRPC()

    def __init__(self, cache, user_data, request_data):
        """Construct a templates read manifest controller."""
        self.ctx = TemplatesReadManifestCtrl.REQUEST_SERIALIZER.load({**user_data, **request_data}, unknown=EXCLUDE)
        super().__init__(cache, user_data, request_data)

    @property
    def context(self):
        """Controller context."""
        return self.ctx

    def template_manifest(self):
        """Reads template manifest."""
        from PIL import Image

        templates_source = fetch_templates_source(source=self.ctx["template_git_url"], reference=self.ctx["ref"])
        manifest = templates_source.manifest.get_raw_content()

        # NOTE: convert icons to base64
        for template in manifest:
            icon = template.get("icon")
            if not icon:
                continue

            icon_path = get_safe_relative_path(path=icon, base=templates_source.path)
            icon = Image.open(templates_source.path / icon_path)
            icon.thumbnail(MAX_ICON_SIZE)

            buffer = BytesIO()
            icon.save(buffer, format="PNG")
            template["icon"] = base64.b64encode(buffer.getvalue())

        return manifest

    def renku_op(self):
        """Renku operation for the controller."""
        # NOTE: We leave it empty since it does not execute renku operation.
        pass

    def to_response(self):
        """Execute controller flow and serialize to service response."""

        return result_response(TemplatesReadManifestCtrl.RESPONSE_SERIALIZER, {"templates": self.template_manifest()})
